import unittest

from cubing_algs.move import Move
from cubing_algs.parsing import parse_moves
from cubing_algs.transform.sign import sign_moves
from cubing_algs.transform.sign import unsign_moves


class TransformSignTestCase(unittest.TestCase):

    def test_unsign_moves(self):
        provide = parse_moves("R' F u' B r")
        expect = parse_moves("R' F Uw' B Rw")

        result = unsign_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_sign_moves(self):
        provide = parse_moves("R' F Uw' B Rw")
        expect = parse_moves("R' F u' B r")

        result = sign_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))



from cubing_algs.constants import CORNER_FACELET_MAP
from cubing_algs.constants import EDGE_FACELET_MAP
from cubing_algs.constants import FACES


def cubies_to_facelets(cp: list[int], co: list[int],
                       ep: list[int], eo: list[int],
                       so: list[int]) -> str:
    """
    Convert Corner/Edge Permutation/Orientation cube state
    to the Kociemba facelets representation string.

    Example - solved state:
      cp = [0, 1, 2, 3, 4, 5, 6, 7]
      co = [0, 0, 0, 0, 0, 0, 0, 0]
      ep = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]
      eo = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
      so = [0, 1, 2, 3, 4, 5]
      facelets = 'UUUUUUUUURRRRRRRRRFFFFFFFFFDDDDDDDDDLLLLLLLLLBBBBBBBBB'

    Example - state after F R moves made:
      cp = [0, 5, 2, 1, 7, 4, 6, 3]
      co = [1, 2, 0, 2, 1, 1, 0, 2]
      ep = [1, 9, 2, 3, 11, 8, 6, 7, 4, 5, 10, 0]
      eo = [1, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0]
      so = [0, 1, 2, 3, 4, 5]
      facelets = 'UUFUUFLLFUUURRRRRRFFRFFDFFDRRBDDBDDBLLDLLDLLDLBBUBBUBB'

    Args:
        cp: Corner Permutation
        co: Corner Orientation
        ep: Edge Permutation
        eo: Edge Orientation
        so: Spatial Orientation

    Returns:
        Cube state in the Kociemba facelets representation string
    """
    facelets = [''] * 54

    for i in range(6):
        facelets[9 * i + 4] = FACES[so[i]]

    for i in range(8):
        for p in range(3):
            real_facelet_idx = CORNER_FACELET_MAP[i][(p + co[i]) % 3]
            color_face_idx = CORNER_FACELET_MAP[cp[i]][p] // 9
            facelets[real_facelet_idx] = FACES[so[color_face_idx]]

    for i in range(12):
        for p in range(2):
            real_facelet_idx = EDGE_FACELET_MAP[i][(p + eo[i]) % 2]
            color_face_idx = EDGE_FACELET_MAP[ep[i]][p] // 9
            facelets[real_facelet_idx] = FACES[so[color_face_idx]]

    return ''.join(facelets)


def facelets_to_cubies(facelets: str) -> tuple[
        list[int], list[int], list[int], list[int], list[int],
]:
    """
    Convert Kociemba facelets representation string to
    Corner/Edge Permutation/Orientation cube state.

    Args:
        facelets: 54-character string representing the cube state
                  in Kociemba facelets format (URFDLB)

    Returns:
        tuple: (cp, co, ep, eo, so) where:
            cp: Corner Permutation list of 8 integers
            co: Corner Orientation list of 8 integers
            ep: Edge Permutation list of 12 integers
            eo: Edge Orientation list of 12 integers
            so: Spatial Orientation list of 6 integers

    Example:
        facelets = 'UUUUUUUUURRRRRRRRRFFFFFFFFFDDDDDDDDDLLLLLLLLLBBBBBBBBB'
        returns: (
            [0, 1, 2, 3, 4, 5, 6, 7],
            [0, 0, 0, 0, 0, 0, 0, 0],
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 1, 2, 3, 4, 5],
        )
    """
    # Get center colors to create color mapping
    so = [0] * 6
    for i in range(6):
        so[i] = FACES.find(facelets[9 * i + 4])

    # Invert the spatial orientation to create color mapping
    so_inv = [0] * 6
    for i in range(6):
        so_inv[so[i]] = i

    # Create color mapping array (convert facelet colors to face indices)
    f = [0] * 54
    for i in range(54):
        f[i] = so_inv[FACES.find(facelets[i])]

    # Initialize arrays
    cp = [0] * 8
    co = [0] * 8
    ep = [0] * 12
    eo = [0] * 12

    # Process corners
    for i in range(8):
        # Find orientation by looking for U or D face (0 or 3 in color mapping)
        ori = 0
        for ori in range(3):
            if f[CORNER_FACELET_MAP[i][ori]] in {0, 3}:
                break

        # Get the other two colors
        col1 = f[CORNER_FACELET_MAP[i][(ori + 1) % 3]]
        col2 = f[CORNER_FACELET_MAP[i][(ori + 2) % 3]]

        # Find matching corner piece
        for j in range(8):
            expected_col1 = CORNER_FACELET_MAP[j][1] // 9
            expected_col2 = CORNER_FACELET_MAP[j][2] // 9
            if col1 == expected_col1 and col2 == expected_col2:
                cp[i] = j
                co[i] = ori % 3
                break

    # Process edges
    for i in range(12):
        color1 = f[EDGE_FACELET_MAP[i][0]]
        color2 = f[EDGE_FACELET_MAP[i][1]]

        for j in range(12):
            expected_color1 = EDGE_FACELET_MAP[j][0] // 9
            expected_color2 = EDGE_FACELET_MAP[j][1] // 9

            if color1 == expected_color1 and color2 == expected_color2:
                ep[i] = j
                eo[i] = 0
                break
            if color1 == expected_color2 and color2 == expected_color1:
                ep[i] = j
                eo[i] = 1
                break

    return cp, co, ep, eo, so



import unittest

from cubing_algs.move import Move
from cubing_algs.parsing import parse_moves
from cubing_algs.transform.auf import remove_auf_moves


class TransformRemoveAUFTestCase(unittest.TestCase):

    def test_remove_auf_moves_pre_one(self):
        provide = parse_moves('U F R B')
        expect = parse_moves('L F R')

        result = remove_auf_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_remove_auf_moves_pre_one_prime(self):
        provide = parse_moves("U' F R B")
        expect = parse_moves('R B L')

        result = remove_auf_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_remove_auf_moves_pre_one_double(self):
        provide = parse_moves('U2 F R B')
        expect = parse_moves('B L F')

        result = remove_auf_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_remove_auf_moves_post_one(self):
        provide = parse_moves('F R B U')
        expect = parse_moves('L F R')

        result = remove_auf_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_remove_auf_moves_post_one_prime(self):
        provide = parse_moves("F R B U'")
        expect = parse_moves('R B L')

        result = remove_auf_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_remove_auf_moves_pre_post_double(self):
        provide = parse_moves('F R B U2')
        expect = parse_moves('B L F')

        result = remove_auf_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_remove_auf_moves_pre_pause_double(self):
        provide = parse_moves('U . U F R B')
        expect = parse_moves('B L F')

        result = remove_auf_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_remove_auf_moves_post_pause_double(self):
        provide = parse_moves('F R B U . U')
        expect = parse_moves('B L F')

        result = remove_auf_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_remove_auf_moves_pre_pause_cancel(self):
        provide = parse_moves("U . U' F R B")
        expect = parse_moves('F R B')

        result = remove_auf_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_remove_auf_moves_post_pause_cancel(self):
        provide = parse_moves("F R B U . U'")
        expect = parse_moves('F R B')

        result = remove_auf_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_remove_auf_moves_double_cancel(self):
        provide = parse_moves('U U F R B U2')
        expect = parse_moves('F R B')

        result = remove_auf_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_remove_auf_moves_pre_post(self):
        provide = parse_moves('U F R B U2')
        expect = parse_moves('R B L')

        result = remove_auf_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_remove_auf_moves_pre_post_paused(self):
        provide = parse_moves('U F R B U . U')
        expect = parse_moves('R B L')

        result = remove_auf_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_remove_auf_moves_empty(self):
        provide = parse_moves('')

        result = remove_auf_moves(provide)

        self.assertEqual(
            result,
            provide,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))



import os
import unittest
from unittest.mock import patch

from cubing_algs.display import TERM_COLORS
from cubing_algs.display import VCubeDisplay
from cubing_algs.vcube import VCube


class TestVCubeDisplay(unittest.TestCase):

    def setUp(self):
        self.cube = VCube()
        self.printer = VCubeDisplay(self.cube)

    def test_init_default_parameters(self):
        printer = VCubeDisplay(self.cube)

        self.assertEqual(printer.cube, self.cube)
        self.assertEqual(printer.cube_size, 3)
        self.assertEqual(printer.face_size, 9)

    @patch.dict(os.environ, {'TERM': 'xterm-256color'})
    def test_display_facelet_with_colors(self):
        with patch('cubing_algs.display.USE_COLORS', True):  # noqa FBT003
            printer = VCubeDisplay(self.cube)
            result = printer.display_facelet('U')
            expected = f"{ TERM_COLORS['white'] } U { TERM_COLORS['reset'] }"
            self.assertEqual(result, expected)

    @patch.dict(os.environ, {'TERM': 'other'})
    def test_display_facelet_without_colors(self):
        with patch('cubing_algs.display.USE_COLORS', False):  # noqa FBT003
            printer = VCubeDisplay(self.cube)
            result = printer.display_facelet('U')
            self.assertEqual(result, ' U ')

    @patch.dict(os.environ, {'TERM': 'xterm-256color'})
    def test_display_facelet_hidden(self):
        with patch('cubing_algs.display.USE_COLORS', True):  # noqa FBT003
            printer = VCubeDisplay(self.cube)
            result = printer.display_facelet('U', '0')
            expected = f"{ TERM_COLORS['hide'] } U { TERM_COLORS['reset'] }"
            self.assertEqual(result, expected)

    def test_display_top_down_face(self):
        printer = VCubeDisplay(self.cube)
        face = 'UUUUUUUUU'

        result = printer.display_top_down_face(face, '111111111')
        lines = result.split('\n')

        self.assertEqual(len(lines), 4)

        for i in range(3):
            line = lines[i]
            self.assertTrue(line.startswith('         '))
            self.assertEqual(line.count('U'), 3)

    def test_display_without_orientation(self):
        printer = VCubeDisplay(self.cube)

        result = printer.display()

        lines = result.split('\n')

        self.assertEqual(len(lines), 10)

        for face in ['U', 'R', 'F', 'D', 'L', 'B']:
            self.assertIn(face, result)

    def test_display_with_orientation(self):
        printer = VCubeDisplay(self.cube)

        initial_state = self.cube.state

        result = printer.display(orientation='DF')
        lines = result.split('\n')

        self.assertEqual(self.cube.state, initial_state)
        self.assertEqual(len(lines), 10)

    def test_display_with_last_layer(self):
        printer = VCubeDisplay(self.cube)

        initial_state = self.cube.state

        result = printer.display(mode='oll')
        lines = result.split('\n')

        self.assertEqual(self.cube.state, initial_state)
        self.assertEqual(len(lines), 6)

    def test_display_f2l(self):
        self.cube.rotate("z2 R U R' U' z2")
        printer = VCubeDisplay(self.cube)

        result = printer.display(mode='f2l')
        lines = result.split('\n')
        self.assertEqual(len(lines), 10)

    def test_display_f2l_initial_no_reorientation(self):
        printer = VCubeDisplay(self.cube)

        result = printer.display(mode='f2l', orientation='UF')
        lines = result.split('\n')
        self.assertEqual(len(lines), 10)

    def test_display_cross(self):
        self.cube.rotate('B L F L F R F L B R')
        printer = VCubeDisplay(self.cube)

        result = printer.display(mode='cross')
        lines = result.split('\n')
        self.assertEqual(len(lines), 10)

    def test_display_structure(self):
        printer = VCubeDisplay(self.cube)
        result = printer.display()

        lines = [line for line in result.split('\n') if line.strip()]

        self.assertEqual(len(lines), 9)

        middle_lines = lines[3:6]
        top_lines = lines[0:3]

        for middle_line in middle_lines:
            for top_line in top_lines:
                self.assertGreater(len(middle_line), len(top_line))

    def test_display_face_order(self):
        cube = VCube()
        printer = VCubeDisplay(cube)

        result = printer.display()
        lines = result.split('\n')

        top_section = ''.join(lines[0:3])
        self.assertIn('U', top_section)
        self.assertNotIn('D', top_section)

        bottom_section = ''.join(lines[6:9])
        self.assertIn('D', bottom_section)
        self.assertNotIn('U', bottom_section)

        middle_section = ''.join(lines[3:6])
        for face in ['L', 'F', 'R', 'B']:
            self.assertIn(face, middle_section)



import unittest

from cubing_algs.exceptions import InvalidBracketError
from cubing_algs.exceptions import InvalidMoveError
from cubing_algs.exceptions import InvalidOperatorError
from cubing_algs.move import Move
from cubing_algs.parsing import check_moves
from cubing_algs.parsing import clean_moves
from cubing_algs.parsing import parse_moves
from cubing_algs.parsing import parse_moves_cfop
from cubing_algs.parsing import split_moves


class CleanMovesTestCase(unittest.TestCase):

    def test_clean_moves(self):
        moves = "R2 L2  (y):F B2' e U R` U’  "  # noqa RUF001
        expect = "R2 L2 y F B2 E U R' U'"
        self.assertEqual(clean_moves(moves), expect)


class SplitMovesTestCase(unittest.TestCase):

    def test_split_moves(self):
        moves = "R2L2yFB2EU'R'U'"
        expect = ['R2', 'L2', 'y', 'F', 'B2', 'E', "U'", "R'", "U'"]
        self.assertEqual(split_moves(moves), expect)

    def test_split_big_moves(self):
        moves = "3R 3Uw' 3b 2-3Dw 3-4d"
        expect = ['3R', "3Uw'", '3b', '2-3Dw', '3-4d']
        self.assertEqual(split_moves(moves), expect)

        moves = "3R3Uw'3b2-3Dw3-4d"
        expect = ['3R', "3Uw'", '3b', '2-3Dw', '3-4d']
        self.assertEqual(split_moves(moves), expect)

    def test_split_timed_moves(self):
        moves = "3R 3Uw'@1500 3b 2-3Dw 3-4d"
        expect = ['3R', "3Uw'@1500", '3b', '2-3Dw', '3-4d']
        self.assertEqual(split_moves(moves), expect)

    def test_split_timed_pauses(self):
        moves = "3R 3Uw'@1500 .@2000 3b 2-3Dw 3-4d"
        expect = ['3R', "3Uw'@1500", '.@2000', '3b', '2-3Dw', '3-4d']
        self.assertEqual(split_moves(moves), expect)

    def test_split_timed_moves_with_pauses(self):
        moves = "3R 3Uw'@1500 . 3b 2-3Dw 3-4d"
        expect = ['3R', "3Uw'@1500", '.', '3b', '2-3Dw', '3-4d']
        self.assertEqual(split_moves(moves), expect)


class CheckMovesTestCase(unittest.TestCase):

    def test_check_moves(self):
        moves = split_moves('R2 L2')
        self.assertTrue(check_moves(moves))

    def test_check_moves_invalid_move(self):
        moves = [Move('T2'), Move('R')]
        self.assertFalse(check_moves(moves))

    def test_check_moves_invalid_wide_standard_move(self):
        moves = [Move('Rw')]
        self.assertTrue(check_moves(moves))
        moves = [Move('Rw3')]
        self.assertFalse(check_moves(moves))
        moves = [Move("Rw2'")]
        self.assertFalse(check_moves(moves))

    def test_check_moves_invalid_wide_sign_move(self):
        moves = [Move('r')]
        self.assertTrue(check_moves(moves))
        moves = [Move('r3')]
        self.assertFalse(check_moves(moves))
        moves = [Move("r2'")]
        self.assertFalse(check_moves(moves))

    def test_check_moves_invalid_modifier(self):
        moves = [Move('R5')]
        self.assertFalse(check_moves(moves))

    def test_check_moves_invalid_too_long(self):
        moves = [Move("R2'")]
        self.assertFalse(check_moves(moves))

    def test_check_moves_invalid_layer(self):
        moves = [Move('2-4R')]
        self.assertFalse(check_moves(moves))


class ParseMovesTestCase(unittest.TestCase):

    def test_parse_moves(self):
        moves = 'R2 L2'
        expect = ['R2', 'L2']
        self.assertEqual(
            parse_moves(moves),
            expect,
        )

    def test_parse_moves_with_pauses(self):
        moves = 'R2 . L2 .'
        expect = ['R2', '.', 'L2', '.']
        self.assertEqual(
            parse_moves(moves),
            expect,
        )

        moves = 'R2 ... L2 .'
        expect = ['R2', '.', '.', '.', 'L2', '.']
        self.assertEqual(
            parse_moves(moves),
            expect,
        )

    def test_parse_list(self):
        moves = ['R2 L2']
        expect = ['R2', 'L2']
        self.assertEqual(
            parse_moves(moves),
            expect,
        )

        moves = ['R2', 'L2']
        expect = ['R2', 'L2']
        self.assertEqual(
            parse_moves(moves),
            expect,
        )

    def test_parse_moves_invalid(self):
        moves = 'R2 T2'
        self.assertRaises(
            InvalidMoveError,
            parse_moves, moves,
            secure=False,
        )

    def test_parse_moves_list_moves(self):
        moves = 'R2 L2'
        expect = ['R2', 'L2']
        self.assertEqual(
            parse_moves(parse_moves(moves)),
            expect,
        )

    def test_parse_moves_algorithm(self):
        moves = 'R2 L2'
        expect = ['R2', 'L2']
        self.assertEqual(
            parse_moves(parse_moves(moves)),
            expect,
        )

    def test_parse_moves_conjugate(self):
        moves = 'F [R, U] F'
        expect = ['F', 'R', 'U', "R'", "U'", 'F']
        self.assertEqual(
            parse_moves(parse_moves(moves)),
            expect,
        )

        moves = 'F[R,U]F'
        self.assertEqual(
            parse_moves(parse_moves(moves)),
            expect,
        )

    def test_parse_moves_conjugate_malformed(self):
        moves = 'F [R, U F'

        self.assertRaises(
            InvalidBracketError,
            parse_moves, moves,
            secure=False,
        )

    def test_parse_moves_conjugate_invalid_moves(self):
        moves = 'F [T, U] F'

        self.assertRaises(
            InvalidMoveError,
            parse_moves, moves,
            secure=False,
        )

        self.assertRaises(
            InvalidMoveError,
            parse_moves, moves,
            secure=True,
        )

    def test_parse_moves_conjugate_nested(self):
        moves = 'F [[R, U], B] F'
        expect = [
            'F',
            'R', 'U', "R'", "U'",
            'B',
            'U', 'R', "U'", "R'",
            "B'",
            'F',
        ]
        self.assertEqual(
            parse_moves(parse_moves(moves)),
            expect,
        )

    def test_parse_moves_commutator(self):
        moves = 'F [R: U] F'
        expect = ['F', 'R', 'U', "R'", 'F']
        self.assertEqual(
            parse_moves(parse_moves(moves)),
            expect,
        )

        moves = 'F[R:U]F'
        self.assertEqual(
            parse_moves(parse_moves(moves)),
            expect,
        )

    def test_parse_moves_commutator_malformed(self):
        moves = 'F [R: U F'

        self.assertRaises(
            InvalidBracketError,
            parse_moves, moves,
            secure=False,
        )

    def test_parse_moves_commutator_invalid_moves(self):
        moves = 'F [T: U] F'

        self.assertRaises(
            InvalidMoveError,
            parse_moves, moves,
            secure=False,
        )

        self.assertRaises(
            InvalidMoveError,
            parse_moves, moves,
            secure=True,
        )

    def test_parse_moves_commutator_nested(self):
        moves = 'F [[R: U]: B] F'
        expect = [
            'F',
            'R', 'U', "R'",
            'B',
            'R', "U'", "R'",
            'F',
        ]
        self.assertEqual(
            parse_moves(parse_moves(moves)),
            expect,
        )

    def test_parse_moves_invalid_operator(self):
        moves = 'F [R; U] F'

        self.assertRaises(
            InvalidOperatorError,
            parse_moves, moves,
            secure=False,
        )

    def test_parse_moves_complex_1(self):
        moves = '[[R: U], D] B [F: [U, R]]'
        expect = [
            'R', 'U', "R'", 'D',
            'R', "U'", "R'", "D'",
            'B',
            'F', 'U', 'R', "U'", "R'", "F'",
        ]
        self.assertEqual(
            parse_moves(parse_moves(moves)),
            expect,
        )

    def test_parse_moves_complex_2(self):
        moves = '[[R F: U L], D] B'
        expect = [
            'R', 'F', 'U', 'L', "F'", "R'",
            'D',
            'R', 'F', "L'", "U'", "F'", "R'",
            "D'",
            'B',
        ]
        self.assertEqual(
            parse_moves(parse_moves(moves)),
            expect,
        )


class ParseMovesCFOPTestCase(unittest.TestCase):

    def test_parse_moves_cfop(self):
        moves = 'R2 L2'
        expect = ['R2', 'L2']
        self.assertEqual(
            parse_moves_cfop(moves),
            expect,
        )

    def test_parse_moves_cfop_cleaned(self):
        moves = 'U R2 L2 y'
        expect = ['R2', 'L2']
        self.assertEqual(
            parse_moves_cfop(moves),
            expect,
        )



from cubing_algs.constants import MAX_ITERATIONS
from cubing_algs.constants import REFAT_MOVES
from cubing_algs.constants import UNFAT_ROTATION_MOVES
from cubing_algs.constants import UNFAT_SLICE_MOVES
from cubing_algs.move import Move


def unfat(
        old_moves: list[Move],
        config: dict[str, list[str]],
) -> list[Move]:
    moves = []

    move_cache: dict[Move, list[Move]] = {}
    for move_str, replacements in config.items():
        move_cache[Move(move_str)] = [Move(m) for m in replacements]

    for move in old_moves:
        move_untimed = move.untimed

        if move_untimed in config:
            if move.is_timed:
                moves.extend(
                    [
                        Move(x + move.time)
                        for x in move_cache[move_untimed]
                    ],
                )
            else:
                moves.extend(move_cache[move_untimed])
        else:
            moves.append(move)

    return moves


def unfat_slice_moves(old_moves: list[Move]) -> list[Move]:
    return unfat(old_moves, UNFAT_SLICE_MOVES)


def unfat_rotation_moves(old_moves: list[Move]) -> list[Move]:
    return unfat(old_moves, UNFAT_ROTATION_MOVES)


def refat(
        old_moves: list[Move],
        config: dict[str, str],
        max_depth: int = MAX_ITERATIONS,
) -> list[Move]:
    if max_depth <= 0:
        return old_moves

    i = 0
    moves = []
    changed = False

    while i < len(old_moves) - 1:
        fatted = f'{ old_moves[i].untimed } { old_moves[i + 1].untimed }'
        if fatted in config:
            moves.append(Move(f'{ config[fatted] }{ old_moves[i].time }'))
            changed = True
            i += 2
        else:
            moves.append(old_moves[i])
            i += 1

    if i < len(old_moves):
        moves.append(old_moves[i])

    if changed:
        return refat(moves, config, max_depth - 1)

    return moves


def refat_moves(old_moves: list[Move]) -> list[Move]:
    return refat(old_moves, REFAT_MOVES)



import unittest

from cubing_algs.move import Move
from cubing_algs.parsing import parse_moves
from cubing_algs.transform.timing import untime_moves


class TransformUntimeTestCase(unittest.TestCase):

    def test_untime_moves(self):
        provide = parse_moves(
            "F@1 R@2 U2@3 F'@4",
        )
        expect = parse_moves("F R U2 F'")

        result = untime_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_untime_big_moves(self):
        provide = parse_moves(
            "2Fw@1 R@2 3U2@3 3f'@4",
        )
        expect = parse_moves("2Fw R 3U2 3f'")

        result = untime_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_untime_moves_untimed(self):
        provide = parse_moves(
            "F@1 R U2 F'@4",
        )
        expect = parse_moves("F R U2 F'")

        result = untime_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))



"""
Move representation for Rubik's cube algorithms.

This module defines the Move class, which represents a single move in a
Rubik's cube algorithm, along with properties to identify move types and
transformations between different notations.
"""
from collections import UserString
from functools import cached_property

from cubing_algs.constants import ALL_BASIC_MOVES
from cubing_algs.constants import DOUBLE_CHAR
from cubing_algs.constants import INNER_MOVES
from cubing_algs.constants import INVERT_CHAR
from cubing_algs.constants import LAYER_SPLIT
from cubing_algs.constants import OUTER_MOVES
from cubing_algs.constants import PAUSE_CHAR
from cubing_algs.constants import ROTATIONS
from cubing_algs.constants import WIDE_CHAR


class Move(UserString):
    """
    Represents a single move in a Rubik's cube algorithm.

    Extends UserString to provide string-like behavior while adding properties
    for move validation and transformation.
    A move consists of an optional layer impacted (such as 2, 3-4),
    a base move (letter) and optional modifiers (such as ', 2, w).

    Examples of valid moves: U, R', F2, Rw, M, x, 3-4Rw, 2F
    """

    # Parsing

    @cached_property
    def layer_move_modifier_time(self) -> tuple[str, str, str, str]:
        layer = ''
        move = ''
        modifier = ''
        time = ''

        layer_match = LAYER_SPLIT.match(self.data)
        if layer_match:
            layer = layer_match.groups()[1]

        kept = self.data[len(layer):]
        if '@' in self.data:
            kept, time = kept.split('@')
            time = f'@{ time }'

        if WIDE_CHAR in kept:
            move, modifier = kept.split(WIDE_CHAR, 1)
            move += WIDE_CHAR
        else:
            move = kept[0]
            modifier = kept[1:]

        return layer, move, modifier, time

    @cached_property
    def layer(self) -> str:
        """
        Extract the layers impacted.
        """
        return self.layer_move_modifier_time[0]

    @cached_property
    def layers(self) -> list[int]:
        """
        List of impacted layers, 0-indexed.
        """
        if not self.layer:
            if self.is_wide_move:
                return [0, 1]
            return [0]
        if '-' not in self.layer:
            if self.is_wide_move:
                return list(range(int(self.layer)))
            return [int(self.layer) - 1]

        start, end = self.layer.split('-', 1)

        return list(range(int(start) - 1, int(end)))

    @cached_property
    def base_move(self) -> str:
        """
        Extract the base letter of the move without modifiers.

        For standard notation, returns the first character.
        For SiGN notation, returns the uppercase of the move.
        """
        if self.is_sign_move:
            return self.raw_base_move[0].upper()
        return self.raw_base_move[0]

    @cached_property
    def raw_base_move(self) -> str:
        """
        Extract the base letter of the move without modifiers
        keeping original notation.

        For standard notation, returns the two first character.
        For SiGN notation, returns the first character.
        """
        return self.layer_move_modifier_time[1]

    @cached_property
    def modifier(self) -> str:
        """
        Extract the modifier part of the move.

        This includes rotation direction (' for counterclockwise) or
        repetition (2 for double moves).
        """
        return self.layer_move_modifier_time[2]

    @cached_property
    def time(self) -> str:
        """
        Extract the time part of the move.
        """
        return self.layer_move_modifier_time[3]

    @cached_property
    def timed(self) -> int | None:
        """
        Integer version of the timed move
        """
        if self.time:
            return int(self.time[1:])
        return None

    # Validation

    @cached_property
    def is_valid_layer(self) -> bool:
        """
        Check if the layer is valid.

        Validates that the layer is effective and valid.
        """
        if not self.is_layered:
            return True

        if '-' in self.layer and not self.is_wide_move:
            return False

        return not len(self.layer.split('-')) > 2

    @cached_property
    def is_valid_move(self) -> bool:
        """
        Check if the base move is valid.

        Validates that the move letter is one of the recognized basic moves.
        """
        if self.is_pause:
            return True

        if WIDE_CHAR in self.data and self.raw_base_move.islower():
            return False

        return self.base_move in ALL_BASIC_MOVES

    @cached_property
    def is_valid_modifier(self) -> bool:
        """
        Check if the modifier is valid.

        Valid modifiers include the invert character (')
        and the double character (2).
        """
        if not self.modifier:
            return True

        if len(self.modifier) > 1:
            return False

        return self.is_double or self.is_counter_clockwise

    @cached_property
    def is_valid(self) -> bool:
        """
        Check if the entire move is valid.

        A move is valid if both its base move and modifier are valid.
        """
        return (
            self.is_valid_layer
            and self.is_valid_move
            and self.is_valid_modifier
        )

    # Move

    @cached_property
    def is_pause(self) -> bool:
        """
        Determine if this is a Pause.

        Pause is a dot (.).
        """
        return self.base_move == PAUSE_CHAR

    @cached_property
    def is_rotation_move(self) -> bool:
        """
        Check if this is a cube rotation move.

        Rotation moves include x, y, and z, which rotate the entire cube.
        """
        return self.base_move in ROTATIONS

    @cached_property
    def is_face_move(self) -> bool:
        """
        Check if this is a face move.

        Face moves are moves that turn a face or slice of the cube,
        as opposed to rotating the entire cube.
        """
        return not self.is_pause and not self.is_rotation_move

    @cached_property
    def is_inner_move(self) -> bool:
        """
        Check if this is an inner slice move.

        Inner slice moves include M, E, and S, which turn the middle slices.
        """
        return self.base_move in INNER_MOVES

    @cached_property
    def is_outer_move(self) -> bool:
        """
        Check if this is an outer face move.

        Outer face moves include U, D, L, R, F, and B,
        which turn the outer faces.
        """
        return self.base_move in OUTER_MOVES

    @cached_property
    def is_wide_move(self) -> bool:
        """
        Check if this is a wide move.

        Wide moves turn two layers at once, such as r, l, u, d, f, and b.
        """
        return self.is_sign_move or WIDE_CHAR in self.raw_base_move

    @cached_property
    def is_layered(self) -> bool:
        """
        Check if this is a layered move.

        Layered moves are moves with layer information, such as 3-4Rw, 3Fw2, u.
        """
        return bool(self.layer) or self.is_wide_move

    @cached_property
    def is_timed(self) -> bool:
        """
        Check if this is a timed move.

        Timed moves are moves with time information, separated by an @.
        """
        return bool(self.time)

    # Notation

    @cached_property
    def is_standard_move(self) -> bool:
        """
        Determine if this move uses Standard notation.
        """
        return not self.is_sign_move

    @cached_property
    def is_sign_move(self) -> bool:
        """
        Determine if this move uses SiGN notation.

        In SiGN notation, wide moves are written with a lower letter
        removing the 'w' (e.g., r instead of Rw).
        """
        if not self.data.islower():
            return False

        return all(char not in self.data for char in [WIDE_CHAR, *ROTATIONS])

    # Modifiers

    @cached_property
    def is_double(self) -> bool:
        """
        Check if this is a double move (180° turn).

        Double moves have a '2' modifier, like U2 or R2.
        """
        return self.modifier == DOUBLE_CHAR

    @cached_property
    def is_clockwise(self) -> bool:
        """
        Check if this is a clockwise move.

        Moves without the invert character (') are clockwise.
        """
        return not self.is_pause and self.modifier != INVERT_CHAR

    @cached_property
    def is_counter_clockwise(self) -> bool:
        """
        Check if this is a counter-clockwise move.

        Moves with the invert character (') are counter-clockwise.
        """
        return not self.is_pause and not self.is_clockwise

    # Transformations

    @cached_property
    def inverted(self) -> 'Move':
        """
        Get the inverted version of this move.

        For a clockwise move, returns the counter-clockwise version.
        For a counter-clockwise move, returns the clockwise version.
        Double moves remain unchanged when inverted.
        """
        if self.is_double or self.is_pause:
            return self

        if self.is_counter_clockwise:
            return Move(
                f'{ self.layer }'
                f'{ self.raw_base_move }'
                f'{ self.time }',
            )
        return Move(
            f'{ self.layer }'
            f'{ self.raw_base_move }'
            f'{ INVERT_CHAR }'
            f'{ self.time }',
        )

    @cached_property
    def doubled(self) -> 'Move':
        """
        Get the doubled version of this move.

        For a single move, returns the double version (180° turn).
        For a double move, returns the single version.
        """
        if self.is_pause:
            return self

        if self.is_double:
            return Move(
                f'{ self.layer }'
                f'{ self.raw_base_move }'
                f'{ self.time }',
            )
        return Move(
            f'{ self.layer }'
            f'{ self.raw_base_move }'
            f'{ DOUBLE_CHAR }'
            f'{ self.time }',
        )

    @cached_property
    def unlayered(self) -> 'Move':
        """
        Convert this move without layer information.

        This converts moves like 3Rw to Rw.
        """
        if self.is_layered:
            return Move(
                f'{ self.raw_base_move }'
                f'{ self.modifier }'
                f'{ self.time }',
            )
        return self

    @cached_property
    def untimed(self) -> 'Move':
        """
        Convert this move without time information.

        This converts moves like 3Rw@200 to 3Rw.
        """
        if self.is_timed:
            return Move(
                f'{ self.layer }'
                f'{ self.raw_base_move }'
                f'{ self.modifier }',
            )
        return self

    @cached_property
    def to_sign(self) -> 'Move':
        """
        Convert this move to SiGN notation.

        In SiGN notation, wide moves use lowercase letters.

        This converts moves like Rw to r.

        This only affects wide moves.
        """
        if self.is_wide_move and not self.is_sign_move:
            return Move(
                f'{ self.layer }'
                f'{ self.base_move.lower() }'
                f'{ self.modifier }'
                f'{ self.time }',
            )
        return self

    @cached_property
    def to_standard(self) -> 'Move':
        """
        Convert this move from SiGN notation to standard notation.

        This converts moves like r to Rw.

        This only affects wide moves.
        """
        if self.is_sign_move:
            return Move(
                f'{ self.layer }'
                f'{ self.base_move.upper() }{ WIDE_CHAR }'
                f'{ self.modifier }'
                f'{ self.time }',
            )
        return self



import unittest

from cubing_algs.move import Move
from cubing_algs.parsing import parse_moves
from cubing_algs.transform.trim import trim_moves


class TransformTrimTestCase(unittest.TestCase):

    def test_trim(self):
        provide = parse_moves('U F R B U')
        expect = parse_moves('F R B')

        result = trim_moves('U')(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_trim_multiple(self):
        provide = parse_moves("U U' F R B U2")
        expect = parse_moves('F R B')

        result = trim_moves('U')(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_trim_multiple_paused(self):
        provide = parse_moves("U' . U2 F R B U . . U'")
        expect = parse_moves('F R B')

        result = trim_moves('U')(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_trim_start(self):
        provide = parse_moves('U F R B U')
        expect = parse_moves('F R B U')

        result = trim_moves('U', end=False)(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_trim_end(self):
        provide = parse_moves('U F R B U')
        expect = parse_moves('U F R B')

        result = trim_moves('U', start=False)(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_trim_empty(self):
        provide = parse_moves('')

        result = trim_moves('U', start=False)(provide)

        self.assertEqual(
            result,
            provide,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))



from itertools import takewhile

from cubing_algs.constants import AUF_CHAR
from cubing_algs.move import Move
from cubing_algs.transform.offset import offset_y2_moves
from cubing_algs.transform.offset import offset_y_moves
from cubing_algs.transform.offset import offset_yprime_moves
from cubing_algs.transform.trim import trim_moves

OFFSET_MAP = {
    1: offset_y_moves,
    2: offset_y2_moves,
    3: offset_yprime_moves,
}


def get_move_score(move: Move) -> int:
    if move.is_double:
        return 2
    if move.is_counter_clockwise:
        return -1
    if move.is_clockwise:
        return 1
    return 0


def is_auf_or_pause(move: Move) -> bool:
    return move.base_move == AUF_CHAR or move.is_pause


def calculate_auf_score(moves: list[Move]) -> int:
    return sum(
        get_move_score(move)
        for move in takewhile(is_auf_or_pause, moves)
    )


def remove_auf_moves(old_moves: list[Move]) -> list[Move]:
    if not old_moves:
        return old_moves

    score = (
        calculate_auf_score(old_moves) +
        calculate_auf_score(list(reversed(old_moves)))
    ) % 4

    transforms = [trim_moves(AUF_CHAR)]
    if score:
        transforms.insert(0, OFFSET_MAP[score])

    return old_moves.transform(*transforms)



"""
Metrics and analysis tools for Rubik's cube algorithms.

This module provides functions to calculate various metrics
for evaluating and comparing Rubik's cube algorithms,
including different move counting systems (HTM, QTM, STM, etc.)
and algorithm composition analysis.
"""
import operator
from typing import Any

from cubing_algs.move import Move

# Dictionary mapping metric names to scoring rules for different move types
MOVE_COUNTS = {
    'htm': {'rotation': [0, 0], 'outer': [1, 0], 'inner': [2, 0]},
    'qtm': {'rotation': [0, 0], 'outer': [0, 1], 'inner': [0, 2]},
    'stm': {'rotation': [0, 0], 'outer': [1, 0], 'inner': [1, 0]},
    'etm': {'rotation': [1, 0], 'outer': [1, 0], 'inner': [1, 0]},
    'rtm': {'rotation': [0, 1], 'outer': [0, 0], 'inner': [0, 0]},
    'qstm': {'rotation': [0, 0], 'outer': [0, 1], 'inner': [0, 1]},
}


def amount(move: Move) -> int:
    """
    Determine the quantity factor for a move.

    Double moves (like U2) count as 2, while single moves count as 1.
    This is used for quarter-turn metrics.
    """
    if move.is_double:
        return 2
    return 1


def move_score(mode: str, field: str,
               moves: list[Move]) -> int:
    """
    Calculate the score for a specific group of moves under a given metric.

    Uses the MOVE_COUNTS dictionary to determine how to score each move
    based on the metric mode and move type.
    """
    datas = MOVE_COUNTS[mode][field]

    return sum(
        datas[0] + (amount(move) * datas[1])
        for move in moves
    )


def compute_score(mode: str,
                  rotations: list[Move],
                  outer: list[Move],
                  inner: list[Move]) -> int:
    """
    Compute the total score for an algorithm under a specific metric.

    Combines scores from all move types (rotations, outer, and inner moves)
    according to the rules of the specified metric.
    """
    return (
        move_score(mode, 'rotation', rotations)
        + move_score(mode, 'outer', outer)
        + move_score(mode, 'inner', inner)
    )


def compute_generators(moves: list[Move]) -> list[str]:
    """
    Identify the most frequently used move faces in an algorithm.

    This function counts how many times each face is turned (ignoring
    direction and whether it's a single or double turn) and returns them in
    order of frequency.
    Rotations are excluded from this analysis.
    """
    count: dict[str, int] = {}
    for move in moves:
        if move.is_rotation_move or move.is_pause:
            continue

        count.setdefault(move.raw_base_move, 0)
        count[move.raw_base_move] += 1

    return [
        k
        for k, v in sorted(
                count.items(),
                key=operator.itemgetter(1),
                reverse=True,
        )
    ]


def regroup_moves(
        moves: list[Move],
) -> tuple[list[Move], list[Move], list[Move], list[Move]]:
    """
    Categorize moves into pause, rotation, outer, and inner move types.

    This separation is necessary for accurate metric calculations, as different
    move types are counted differently depending on the metric.
    """
    pauses = []
    rotations = []
    outer_moves = []
    inner_moves = []

    for move in moves:
        if move.is_pause:
            pauses.append(move)
        elif move.is_outer_move:
            outer_moves.append(move)
        elif move.is_inner_move:
            inner_moves.append(move)
        else:
            rotations.append(move)

    return pauses, rotations, outer_moves, inner_moves


def compute_metrics(moves: list[Move]) -> dict[str, Any]:
    """
    Calculate a comprehensive set of metrics for an algorithm.

    This function computes various metrics including:
    - Move counts by type (rotations, outer moves, inner moves)
    - Standard metrics (HTM, QTM, STM, ETM, QSTM)
    - Generator analysis (most used faces)

    Returns:
        dict[str, Any]: A dictionary containing all calculated metrics.
        Keys include:
            - rotations: Number of rotation moves
            - outer_moves: Number of outer face moves
            - inner_moves: Number of inner slice moves
            - htm: Half Turn Metric score
            - qtm: Quarter Turn Metric score
            - stm: Slice Turn Metric score
            - etm: Execution Turn Metric score
            - rtm: Rotation Turn Metric score
            - qstm: Quarter Slice Turn Metric score
            - generators: List of most frequently used faces
    """
    pauses, rotations, outer_moves, inner_moves = regroup_moves(moves)

    return {
        'pauses': len(pauses),
        'rotations': len(rotations),
        'outer_moves': len(outer_moves),
        'inner_moves': len(inner_moves),
        'htm': compute_score('htm', rotations, outer_moves, inner_moves),
        'qtm': compute_score('qtm', rotations, outer_moves, inner_moves),
        'stm': compute_score('stm', rotations, outer_moves, inner_moves),
        'etm': compute_score('etm', rotations, outer_moves, inner_moves),
        'rtm': compute_score('rtm', rotations, outer_moves, inner_moves),
        'qstm': compute_score('qstm', rotations, outer_moves, inner_moves),
        'generators': compute_generators(moves),
    }



import unittest

from cubing_algs.move import Move


class MoveTestCase(unittest.TestCase):

    def test_base_move(self):
        self.assertEqual(Move('U').base_move, 'U')
        self.assertEqual(Move('x2').base_move, 'x')
        self.assertEqual(Move('Uw').base_move, 'U')
        self.assertEqual(Move('u').base_move, 'U')
        self.assertEqual(Move('.').base_move, '.')
        self.assertEqual(Move("2-4u'@200").base_move, 'U')
        self.assertEqual(Move("2-4Uw'@200").base_move, 'U')

    def test_raw_base_move(self):
        self.assertEqual(Move('U').raw_base_move, 'U')
        self.assertEqual(Move('x2').raw_base_move, 'x')
        self.assertEqual(Move('Uw').raw_base_move, 'Uw')
        self.assertEqual(Move('u').raw_base_move, 'u')
        self.assertEqual(Move("u'").raw_base_move, 'u')
        self.assertEqual(Move("2-4u'@200").raw_base_move, 'u')
        self.assertEqual(Move("2-4Uw'@200").raw_base_move, 'Uw')
        self.assertEqual(Move('.').raw_base_move, '.')

    def test_modifier(self):
        self.assertEqual(Move('U').modifier, '')
        self.assertEqual(Move('x2').modifier, '2')
        self.assertEqual(Move('Uw').modifier, '')
        self.assertEqual(Move('Uw2').modifier, '2')
        self.assertEqual(Move('.').modifier, '')

    def test_is_valid(self):
        self.assertTrue(Move('U').is_valid)
        self.assertTrue(Move('u').is_valid)
        self.assertTrue(Move('u2').is_valid)
        self.assertTrue(Move('Uw').is_valid)
        self.assertFalse(Move('T').is_valid)
        self.assertFalse(Move('uw').is_valid)
        self.assertFalse(Move('Ux').is_valid)
        self.assertFalse(Move("U2'").is_valid)
        self.assertFalse(Move('3-4R').is_valid)
        self.assertTrue(Move('3-4Rw').is_valid)
        self.assertTrue(Move('2Dw2').is_valid)
        self.assertTrue(Move('2Dw2@200').is_valid)
        self.assertTrue(Move('.').is_valid)

    def test_is_valid_move(self):
        self.assertTrue(Move('U').is_valid_move)
        self.assertTrue(Move('u').is_valid_move)
        self.assertTrue(Move('Uw').is_valid_move)
        self.assertFalse(Move('T').is_valid_move)
        self.assertFalse(Move('uw').is_valid_move)
        self.assertTrue(Move('.').is_valid_move)

    def test_is_valid_modifier(self):
        self.assertTrue(Move('U').is_valid_modifier)
        self.assertTrue(Move('U2').is_valid_modifier)
        self.assertTrue(Move("U'").is_valid_modifier)
        self.assertTrue(Move('Uw2').is_valid_modifier)
        self.assertTrue(Move("Uw'").is_valid_modifier)
        self.assertFalse(Move("U2'").is_valid_modifier)
        self.assertFalse(Move("Uw2'").is_valid_modifier)
        self.assertTrue(Move('.').is_valid_modifier)

    def test_is_valid_layer(self):
        self.assertTrue(Move('3Rw').is_valid_layer)
        self.assertTrue(Move('3R').is_valid_layer)
        self.assertTrue(Move('3-4Rw').is_valid_layer)
        self.assertTrue(Move('3-4r').is_valid_layer)
        self.assertFalse(Move('3-4R').is_valid_layer)
        self.assertFalse(Move('2-3-4R').is_valid_layer)
        self.assertTrue(Move('2Dw2').is_valid_layer)
        self.assertTrue(Move('3-4Rw').is_valid_layer)
        self.assertTrue(Move('.').is_valid_layer)

    def test_is_double(self):
        self.assertFalse(Move('U').is_double)
        self.assertFalse(Move("U'").is_double)
        self.assertTrue(Move('U2').is_double)
        self.assertFalse(Move('.').is_double)

    def test_is_clockwise(self):
        self.assertTrue(Move('U').is_clockwise)
        self.assertFalse(Move("U'").is_clockwise)
        self.assertFalse(Move('.').is_clockwise)

    def test_is_counter_clockwise(self):
        self.assertTrue(Move("U'").is_counter_clockwise)
        self.assertFalse(Move('U').is_counter_clockwise)
        self.assertFalse(Move('.').is_counter_clockwise)

    def test_is_pause(self):
        self.assertFalse(Move('U').is_pause)
        self.assertFalse(Move('x').is_pause)
        self.assertTrue(Move('.').is_pause)
        self.assertTrue(Move('.@100').is_pause)

    def test_is_rotation_move(self):
        self.assertFalse(Move('U').is_rotation_move)
        self.assertFalse(Move('.').is_rotation_move)
        self.assertTrue(Move('x').is_rotation_move)

    def test_is_face_move(self):
        self.assertFalse(Move('x').is_face_move)
        self.assertFalse(Move('.').is_face_move)
        self.assertTrue(Move('F').is_face_move)

    def test_is_inner_move(self):
        self.assertFalse(Move('R').is_inner_move)
        self.assertTrue(Move('M').is_inner_move)
        self.assertFalse(Move('.').is_inner_move)

    def test_is_outer_move(self):
        self.assertFalse(Move('x').is_outer_move)
        self.assertFalse(Move('M').is_outer_move)
        self.assertTrue(Move('R').is_outer_move)
        self.assertTrue(Move('r2').is_outer_move)
        self.assertFalse(Move('.').is_outer_move)

    def test_is_wide_move(self):
        self.assertFalse(Move('x').is_wide_move)
        self.assertFalse(Move('R').is_wide_move)
        self.assertTrue(Move('r2').is_wide_move)
        self.assertTrue(Move('rw2').is_wide_move)
        self.assertTrue(Move('Rw2').is_wide_move)
        self.assertFalse(Move('.').is_wide_move)

    def test_is_layered(self):
        self.assertFalse(Move('x').is_layered)
        self.assertFalse(Move('R').is_layered)
        self.assertFalse(Move('.').is_layered)
        self.assertTrue(Move('2R').is_layered)
        self.assertTrue(Move('2Rw').is_layered)
        self.assertTrue(Move('2r').is_layered)
        self.assertTrue(Move('4Rw').is_layered)
        self.assertTrue(Move('3-4Rw').is_layered)
        self.assertTrue(Move('r2').is_layered)

    def test_is_timed(self):
        self.assertFalse(Move('x').is_timed)
        self.assertFalse(Move('R').is_timed)
        self.assertFalse(Move('.').is_timed)
        self.assertTrue(Move('x@500').is_timed)
        self.assertTrue(Move('R@500').is_timed)
        self.assertTrue(Move('2R@500').is_timed)
        self.assertTrue(Move('2Rw@500').is_timed)
        self.assertTrue(Move('2r@500').is_timed)
        self.assertTrue(Move('4Rw@500').is_timed)
        self.assertTrue(Move('3-4Rw@500').is_timed)
        self.assertTrue(Move('r2@500').is_timed)
        self.assertTrue(Move('.@500').is_timed)

    def test_is_standard_move(self):
        self.assertTrue(Move('x').is_standard_move)
        self.assertTrue(Move('R').is_standard_move)
        self.assertTrue(Move('.').is_standard_move)
        self.assertTrue(Move('uw').is_standard_move)
        self.assertTrue(Move('xw').is_standard_move)
        self.assertTrue(Move('xw2').is_standard_move)
        self.assertTrue(Move('Rw').is_standard_move)
        self.assertTrue(Move('Rw2').is_standard_move)
        self.assertFalse(Move('u2').is_standard_move)
        self.assertFalse(Move("u'").is_standard_move)
        self.assertFalse(Move("u'@100").is_standard_move)
        self.assertFalse(Move("2-4u'@100").is_standard_move)

    def test_is_sign_move(self):
        self.assertFalse(Move('x').is_sign_move)
        self.assertFalse(Move('R').is_sign_move)
        self.assertFalse(Move('.').is_sign_move)
        self.assertFalse(Move('uw').is_sign_move)
        self.assertFalse(Move('xw').is_sign_move)
        self.assertFalse(Move('xw2').is_sign_move)
        self.assertFalse(Move('Rw').is_sign_move)
        self.assertFalse(Move('Rw2').is_sign_move)
        self.assertTrue(Move('u2').is_sign_move)
        self.assertTrue(Move("u'").is_sign_move)
        self.assertTrue(Move("u'@100").is_sign_move)
        self.assertTrue(Move("2-4u'@100").is_sign_move)

    def test_inverted(self):
        self.assertEqual(Move('R').inverted, Move("R'"))
        self.assertEqual(Move("R'").inverted, Move('R'))
        self.assertEqual(Move("x'").inverted, Move('x'))
        self.assertEqual(Move('R2').inverted, Move('R2'))
        self.assertEqual(Move('3R2').inverted, Move('3R2'))
        self.assertEqual(Move('3R@200').inverted, Move("3R'@200"))
        self.assertEqual(Move('.').inverted, Move('.'))

    def test_doubled(self):
        self.assertEqual(Move('R').doubled, Move('R2'))
        self.assertEqual(Move("R'").doubled, Move('R2'))
        self.assertEqual(Move("x'").doubled, Move('x2'))
        self.assertEqual(Move('R2').doubled, Move('R'))
        self.assertEqual(Move('3R2').doubled, Move('3R'))
        self.assertEqual(Move('3R2@200').doubled, Move('3R@200'))
        self.assertEqual(Move('.').doubled, Move('.'))

    def test_to_sign(self):
        self.assertEqual(Move('R').to_sign, Move('R'))
        self.assertEqual(Move('x').to_sign, Move('x'))
        self.assertEqual(Move('r').to_sign, Move('r'))
        self.assertEqual(Move('r2').to_sign, Move('r2'))
        self.assertEqual(Move('r@200').to_sign, Move('r@200'))
        self.assertEqual(Move('.').to_sign, Move('.'))

    def test_to_standard(self):
        self.assertEqual(Move('R').to_standard, Move('R'))
        self.assertEqual(Move('x').to_standard, Move('x'))
        self.assertEqual(Move('r').to_standard, Move('Rw'))
        self.assertEqual(Move('r2').to_standard, Move('Rw2'))
        self.assertEqual(Move('r@200').to_standard, Move('Rw@200'))
        self.assertEqual(Move('.').to_standard, Move('.'))

    def test_unlayered(self):
        self.assertEqual(Move('R').unlayered, Move('R'))
        self.assertEqual(Move('2R').unlayered, Move('R'))
        self.assertEqual(Move('2-4Rw').unlayered, Move('Rw'))
        self.assertEqual(Move('2-4r').unlayered, Move('r'))
        self.assertEqual(Move('Rw').unlayered, Move('Rw'))
        self.assertEqual(Move('4Rw').unlayered, Move('Rw'))
        self.assertEqual(Move('u').unlayered, Move('u'))
        self.assertEqual(Move('2u').unlayered, Move('u'))
        self.assertEqual(Move('4u@200').unlayered, Move('u@200'))
        self.assertEqual(Move('.').unlayered, Move('.'))

    def test_untimed(self):
        self.assertEqual(Move('R').untimed, Move('R'))
        self.assertEqual(Move('2R').untimed, Move('2R'))
        self.assertEqual(Move('2-4Rw').untimed, Move('2-4Rw'))
        self.assertEqual(Move('R@100').untimed, Move('R'))
        self.assertEqual(Move('2R@100').untimed, Move('2R'))
        self.assertEqual(Move('2-4Rw@100').untimed, Move('2-4Rw'))
        self.assertEqual(Move('.@100').untimed, Move('.'))

    def test_layer(self):
        self.assertEqual(Move('R').layer, '')
        self.assertEqual(Move('.').layer, '')

        self.assertEqual(Move('2R').layer, '2')

        self.assertEqual(Move('2Rw').layer, '2')
        self.assertEqual(Move('2r').layer, '2')

        self.assertEqual(Move('3Rw').layer, '3')
        self.assertEqual(Move('3r').layer, '3')

        self.assertEqual(Move('3-4Rw').layer, '3-4')
        self.assertEqual(Move('3-4r').layer, '3-4')

        self.assertEqual(Move('1-3-4r').layer, '1-3-4')

        self.assertEqual(Move('2Dw2').layer, '2')

    def test_big_moves_standard(self):
        move = Move('3Rw')

        self.assertEqual(move.layer, '3')
        self.assertEqual(move.to_sign, Move('3r'))
        self.assertEqual(move.to_standard, Move('3Rw'))
        self.assertEqual(move.doubled, Move('3Rw2'))
        self.assertEqual(move.inverted, Move("3Rw'"))
        self.assertEqual(move.unlayered, Move('Rw'))
        self.assertEqual(move.raw_base_move, 'Rw')
        self.assertEqual(move.base_move, 'R')
        self.assertEqual(move.modifier, '')
        self.assertFalse(move.is_sign_move)
        self.assertFalse(move.is_timed)
        self.assertTrue(move.is_layered)
        self.assertTrue(move.is_wide_move)
        self.assertTrue(move.is_outer_move)
        self.assertFalse(move.is_inner_move)
        self.assertTrue(move.is_face_move)
        self.assertFalse(move.is_rotation_move)
        self.assertTrue(move.is_clockwise)
        self.assertFalse(move.is_counter_clockwise)
        self.assertFalse(move.is_double)

    def test_big_moves_sign(self):
        move = Move('3r')

        self.assertEqual(move.layer, '3')
        self.assertEqual(move.to_sign, Move('3r'))
        self.assertEqual(move.to_standard, Move('3Rw'))
        self.assertEqual(move.doubled, Move('3r2'))
        self.assertEqual(move.inverted, Move("3r'"))
        self.assertEqual(move.unlayered, Move('r'))
        self.assertEqual(move.raw_base_move, 'r')
        self.assertEqual(move.base_move, 'R')
        self.assertEqual(move.modifier, '')
        self.assertTrue(move.is_sign_move)
        self.assertFalse(move.is_timed)
        self.assertTrue(move.is_layered)
        self.assertTrue(move.is_wide_move)
        self.assertTrue(move.is_outer_move)
        self.assertFalse(move.is_inner_move)
        self.assertTrue(move.is_face_move)
        self.assertFalse(move.is_rotation_move)
        self.assertTrue(move.is_clockwise)
        self.assertFalse(move.is_counter_clockwise)
        self.assertFalse(move.is_double)

    def test_layers(self):
        self.assertEqual(Move('R').layers, [0])
        self.assertEqual(Move('2R').layers, [1])

        self.assertEqual(Move('Rw').layers, [0, 1])
        self.assertEqual(Move('r').layers, [0, 1])

        self.assertEqual(Move('2Rw').layers, [0, 1])
        self.assertEqual(Move('2r').layers, [0, 1])

        self.assertEqual(Move('3Rw').layers, [0, 1, 2])
        self.assertEqual(Move('3r').layers, [0, 1, 2])

        self.assertEqual(Move('3-4Rw').layers, [2, 3])
        self.assertEqual(Move('3-4r').layers, [2, 3])

        self.assertEqual(Move('2-4r').layers, [1, 2, 3])

        self.assertEqual(Move('2Dw2').layers, [0, 1])

    def test_timed(self):
        self.assertEqual(Move('R').time, '')
        self.assertEqual(Move('R').timed, None)

        self.assertEqual(Move('.').time, '')
        self.assertEqual(Move('.').timed, None)

        self.assertEqual(Move('R@1500').time, '@1500')
        self.assertEqual(Move('R@1500').timed, 1500)

        self.assertEqual(Move('2-4r@200').time, '@200')
        self.assertEqual(Move('2-4r@200').timed, 200)

        self.assertEqual(Move('.@100').time, '@100')
        self.assertEqual(Move('.@100').timed, 100)



import unittest

from cubing_algs.move import Move
from cubing_algs.parsing import parse_moves
from cubing_algs.transform.slice import reslice
from cubing_algs.transform.slice import reslice_e_moves
from cubing_algs.transform.slice import reslice_e_timed_moves
from cubing_algs.transform.slice import reslice_m_moves
from cubing_algs.transform.slice import reslice_m_timed_moves
from cubing_algs.transform.slice import reslice_moves
from cubing_algs.transform.slice import reslice_s_moves
from cubing_algs.transform.slice import reslice_s_timed_moves
from cubing_algs.transform.slice import reslice_timed_moves
from cubing_algs.transform.slice import unslice_rotation_moves
from cubing_algs.transform.slice import unslice_wide_moves


class TransformSliceTestCase(unittest.TestCase):

    def test_unslice_rotation_moves(self):
        provide = parse_moves('M2 U S E')
        expect = parse_moves("L2 R2 x2 U F' B z D' U y'")

        result = unslice_rotation_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_unslice_wide_moves(self):
        provide = parse_moves('M2 U S E')
        expect = parse_moves("r2 R2 U f F' u' U")

        result = unslice_wide_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_unslice_timed_moves(self):
        provide = parse_moves('M2@1 U@2 S@3 E@4')
        expect = parse_moves(
            "L2@1 R2@1 x2@1 U@2 F'@3 B@3 z@3 D'@4 U@4 y'@4",
        )

        result = unslice_rotation_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_unslice_timed_moves_pauses(self):

        provide = parse_moves('M2@1 .@2 U@3 S@4 E@5')
        expect = parse_moves(
            "L2@1 R2@1 x2@1 .@2 U@3 F'@4 B@4 z@4 D'@5 U@5 y'@5",
        )

        result = unslice_rotation_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_reslice_moves(self):
        provide = parse_moves("U' D")
        expect = parse_moves("E' y'")

        result = reslice_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_reslice_moves_alt(self):
        provide = parse_moves("D U'")
        expect = parse_moves("E' y'")

        result = reslice_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_reslice_moves_fat(self):
        provide = parse_moves("r' R")
        expect = parse_moves('M')

        result = reslice_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_reslice_moves_fat_alt(self):
        provide = parse_moves("R r'")
        expect = parse_moves('M')

        result = reslice_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_reslice_e_moves(self):
        provide = parse_moves("U' D F")
        expect = parse_moves("E' y' F")

        result = reslice_e_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_reslice_m_moves(self):
        provide = parse_moves("L' R F")
        expect = parse_moves('M x F')

        result = reslice_m_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_reslice_m_moves_double(self):
        provide = parse_moves('R2 L2 F')
        expect = parse_moves('M2 x2 F')

        result = reslice_m_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_reslice_m_moves_timed(self):
        provide = parse_moves("L'@100 R@200 F@300")
        expect = parse_moves('M@100 x@100 F@300')

        result = reslice_m_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_reslice_m_moves_big(self):
        provide = parse_moves("L' R 2F")
        expect = parse_moves('M x 2F')

        result = reslice_m_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

        provide = parse_moves("2L' R 2F")
        expect = parse_moves("2L' R 2F")

        result = reslice_m_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_reslice_m_moves_big_timed(self):
        provide = parse_moves("L' R 2F@200")
        expect = parse_moves('M x 2F@200')

        result = reslice_m_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

        provide = parse_moves("2L'@100 R@200 2F@300")
        expect = parse_moves("2L'@100 R@200 2F@300")

        result = reslice_m_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_reslice_m_moves_big_timed_pauses(self):
        provide = parse_moves("L' . R 2F@200")
        expect = parse_moves("L' . R 2F@200")

        result = reslice_m_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_reslice_s_moves(self):
        provide = parse_moves("B' F F")
        expect = parse_moves("S' z F")

        result = reslice_s_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_reslice_max(self):
        provide = parse_moves("U' D")

        self.assertEqual(
            reslice(provide, {}, 0),
            provide,
        )


class TransformSliceTimedTestCase(unittest.TestCase):

    def test_reslice_timed_moves(self):
        provide = parse_moves("U'@100 D@150")
        expect = parse_moves("E'@100 y'@100")

        result = reslice_timed_moves(50)(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_reslice_timed_moves_failed(self):
        provide = parse_moves("U'@100 D@150")

        result = reslice_timed_moves(10)(provide)

        self.assertEqual(
            result,
            provide,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_reslice_timed_moves_chained(self):
        provide = parse_moves(
            "F@21031 "
            "B'@23249 F@23279 "
            "B'@23520 F@23520 "
            "D@23789 "
            "R@24060 L'@24060 "
            "L'@24300 R@24301 "
            "U'@24809 R'@25499 "
            "L@25529 D@26309 "
            "U'@26311 U'@26639 "
            "D@26640 L@27089 "
            "R'@27090 D@27780",
        )
        expect = parse_moves(
            "F@21031 "
            "B'@23249 F@23279 "
            "S'@23520 z@23520 "
            "D@23789 "
            "M@24060 x@24060 "
            "M@24300 x@24300 "
            "U'@24809 R'@25499 "
            "L@25529 "
            "E'@26309 y'@26309 "
            "E'@26639 y'@26639 "
            "M'@27089 x'@27089 "
            "D@27780",
        )

        result = reslice_timed_moves(20)(provide)
        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

        expect = parse_moves(
            "F@21031 "
            "S'@23249 z@23249 "
            "S'@23520 z@23520 "
            "D@23789 "
            "M@24060 x@24060 "
            "M@24300 x@24300 "
            "U'@24809 "
            "M'@25499 x'@25499 "
            "E'@26309 y'@26309 "
            "E'@26639 y'@26639 "
            "M'@27089 x'@27089 "
            "D@27780",
        )

        result = reslice_timed_moves(50)(provide)
        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_reslice_timed_moves_without_time(self):
        provide = parse_moves("U' D")
        expect = parse_moves("E'y'")

        result = reslice_timed_moves()(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_reslice_m_timed_moves(self):
        provide = parse_moves("L'@0 R@30 F@70")
        expect = parse_moves('M@0 x@0 F@70')

        result = reslice_m_timed_moves(50)(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_reslice_s_timed_moves(self):
        provide = parse_moves("B'@0 F@30 F@70")
        expect = parse_moves("S'@0 z@0 F@70")

        result = reslice_s_timed_moves(50)(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_reslice_e_timed_moves(self):
        provide = parse_moves("U'@0 D@30 F@70")
        expect = parse_moves("E'@0 y'@0 F@70")

        result = reslice_e_timed_moves(50)(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))



import unittest

from cubing_algs.facelets import cubies_to_facelets
from cubing_algs.facelets import facelets_to_cubies


class CubiesToFaceletsTestCase(unittest.TestCase):

    def test_cubies_to_facelets_solved(self):
        cp = [0, 1, 2, 3, 4, 5, 6, 7]
        co = [0, 0, 0, 0, 0, 0, 0, 0]
        ep = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]
        eo = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        so = [0, 1, 2, 3, 4, 5]
        facelets = 'UUUUUUUUURRRRRRRRRFFFFFFFFFDDDDDDDDDLLLLLLLLLBBBBBBBBB'

        self.assertEqual(
            cubies_to_facelets(
                cp, co,
                ep, eo,
                so,
            ),
            facelets,
        )

    def test_cubies_to_facelets_solved_oriented(self):
        cp = [0, 1, 2, 3, 4, 5, 6, 7]
        co = [0, 0, 0, 0, 0, 0, 0, 0]
        ep = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]
        eo = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        so = [3, 4, 2, 0, 1, 5]
        facelets = (
            'DDDDDDDDD'
            'LLLLLLLLL'
            'FFFFFFFFF'
            'UUUUUUUUU'
            'RRRRRRRRR'
            'BBBBBBBBB'
        )

        self.assertEqual(
            cubies_to_facelets(
                cp, co,
                ep, eo,
                so,
            ),
            facelets,
        )

    def test_cubies_to_facelets(self):
        cp = [0, 5, 2, 1, 7, 4, 6, 3]
        co = [1, 2, 0, 2, 1, 1, 0, 2]
        ep = [1, 9, 2, 3, 11, 8, 6, 7, 4, 5, 10, 0]
        eo = [1, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0]
        so = [0, 1, 2, 3, 4, 5]
        facelets = 'UUFUUFLLFUUURRRRRRFFRFFDFFDRRBDDBDDBLLDLLDLLDLBBUBBUBB'

        self.assertEqual(
            cubies_to_facelets(
                cp, co,
                ep, eo,
                so,
            ),
            facelets,
        )

    def test_cubies_to_facelets_oriented(self):
        cp = [4, 0, 1, 3, 7, 5, 6, 2]
        co = [2, 0, 0, 1, 1, 0, 0, 2]
        ep = [8, 0, 1, 2, 11, 5, 6, 7, 4, 9, 10, 3]
        eo = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        so = [2, 1, 3, 5, 4, 0]
        facelets = 'FFRFFDFFDRRURRURRURRBDDBDDBBBUBBUBBLDDDLLLLLLFLLFUUFUU'

        self.assertEqual(
            cubies_to_facelets(
                cp, co,
                ep, eo,
                so,
            ),
            facelets,
        )


class FaceletsToCubiesTestCase(unittest.TestCase):

    def test_facelets_to_cubies_solved(self):
        cp = [0, 1, 2, 3, 4, 5, 6, 7]
        co = [0, 0, 0, 0, 0, 0, 0, 0]
        ep = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]
        eo = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        so = [0, 1, 2, 3, 4, 5]
        facelets = 'UUUUUUUUURRRRRRRRRFFFFFFFFFDDDDDDDDDLLLLLLLLLBBBBBBBBB'

        self.assertEqual(
            facelets_to_cubies(facelets),
            (
                cp, co,
                ep, eo,
                so,
            ),
        )

    def test_facelets_to_cubies(self):
        cp = [0, 5, 2, 1, 7, 4, 6, 3]
        co = [1, 2, 0, 2, 1, 1, 0, 2]
        ep = [1, 9, 2, 3, 11, 8, 6, 7, 4, 5, 10, 0]
        eo = [1, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0]
        so = [0, 1, 2, 3, 4, 5]
        facelets = 'UUFUUFLLFUUURRRRRRFFRFFDFFDRRBDDBDDBLLDLLDLLDLBBUBBUBB'

        self.assertEqual(
            facelets_to_cubies(facelets),
            (
                cp, co,
                ep, eo,
                so,
            ),
        )

    def test_facelets_to_cubies_oriented(self):
        cp = [4, 0, 1, 3, 7, 5, 6, 2]
        co = [2, 0, 0, 1, 1, 0, 0, 2]
        ep = [8, 0, 1, 2, 11, 5, 6, 7, 4, 9, 10, 3]
        eo = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        so = [2, 1, 3, 5, 4, 0]
        facelets = 'FFRFFDFFDRRURRURRURRBDDBDDBBBUBBUBBLDDDLLLLLLFLLFUUFUU'

        self.assertEqual(
            facelets_to_cubies(facelets),
            (
                cp, co,
                ep, eo,
                so,
            ),
        )



from collections.abc import Callable
from itertools import dropwhile

from cubing_algs.move import Move


def trim_moves(
        trim_move: str,
        start: bool = True, end: bool = True,  # noqa: FBT001, FBT002
) -> Callable[[list[Move]], list[Move]]:

    def trimmer(old_moves: list[Move]) -> list[Move]:
        if not old_moves:
            return []

        moves = old_moves.copy()

        def should_trim(m: Move) -> bool:
            return m.base_move == trim_move or m.is_pause

        if start:
            moves = list(
                dropwhile(should_trim, moves),
            )

        if end:
            moves = list(
                reversed(
                    list(
                        dropwhile(should_trim, reversed(moves)),
                    ),
                ),
            )

        return moves

    return trimmer



import unittest

from cubing_algs.move import Move
from cubing_algs.parsing import parse_moves
from cubing_algs.transform.mirror import mirror_moves


class TransformMirrorTestCase(unittest.TestCase):

    def test_mirror_moves(self):
        provide = parse_moves(
            "F R U2 F'",
        )
        expect = parse_moves("F U2 R' F'")

        result = mirror_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_mirror_big_moves(self):
        provide = parse_moves(
            "2Fw R 3U2 3f'",
        )
        expect = parse_moves("3f 3U2 R' 2Fw'")

        result = mirror_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_timed_moves(self):
        provide = parse_moves(
            "F@1 R@2 U2@3 F'@4",
        )
        expect = parse_moves("F@4 U2@3 R'@2 F'@1")

        result = mirror_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_timed_moves_with_pauses(self):
        provide = parse_moves(
            "F@1 .@2 R@3 U2@4 F'@5",
        )
        expect = parse_moves("F@5 U2@4 R'@3 .@2 F'@1")

        result = mirror_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))



from cubing_algs.algorithm import Algorithm
from cubing_algs.exceptions import InvalidBracketError
from cubing_algs.exceptions import InvalidOperatorError
from cubing_algs.transform.mirror import mirror_moves


def find_innermost_brackets(text: str) -> tuple[int, int] | None:
    """
    Find the position of the innermost (deepest nested) brackets.
    """
    max_depth = 0
    current_depth = 0
    innermost_start = -1
    innermost_end = -1

    i = 0
    while i < len(text):
        if text[i] == '[':
            current_depth += 1
            if current_depth > max_depth:  # pragma: no branch
                max_depth = current_depth
                innermost_start = i
        elif text[i] == ']':
            if current_depth == max_depth and innermost_start != -1:
                innermost_end = i
                return (innermost_start, innermost_end)
            current_depth -= 1
        i += 1

    return None


def split_on_separator(text: str, separator: str) -> tuple[str, str] | None:
    """
    Split text on separator, but only if the separator is at the top level
    (not inside nested brackets).
    """
    bracket_depth = 0

    for i, char in enumerate(text):
        if char == '[':
            bracket_depth += 1
        elif char == ']':
            bracket_depth -= 1
        elif char == separator and bracket_depth == 0:
            return (text[:i], text[i + 1:])

    return None


def invert_moves(old_moves: str) -> str:
    """
    Invert an algorithm string (reverse order and invert each move).
    """
    algo = Algorithm() + old_moves

    return str(algo.transform(mirror_moves))


def expand_commutators_and_conjugates(moves: str) -> str:
    """
    Expand commutators [A, B] and conjugates [A: B] in a move string.

    Commutator [A, B] = A B A' B'
    Conjugate [A: B] = A B A'
    """
    result = moves

    while '[' in result:
        bracket_pos = find_innermost_brackets(result)
        if bracket_pos is None:
            msg = f'Malformed bracket in { result }'
            raise InvalidBracketError(msg)

        start, end = bracket_pos

        bracket_content = result[start + 1:end]

        colon_split = split_on_separator(bracket_content, ':')
        comma_split = split_on_separator(bracket_content, ',')

        if colon_split is not None:
            a_part, b_part = colon_split

            a_expanded = expand_commutators_and_conjugates(a_part)
            b_expanded = expand_commutators_and_conjugates(b_part)

            a_inverted = invert_moves(a_expanded)
            expanded = f'{a_expanded} {b_expanded} {a_inverted}'

        elif comma_split is not None:
            a_part, b_part = comma_split

            a_expanded = expand_commutators_and_conjugates(a_part)
            b_expanded = expand_commutators_and_conjugates(b_part)

            a_inverted = invert_moves(a_expanded)
            b_inverted = invert_moves(b_expanded)
            expanded = f'{a_expanded} {b_expanded} {a_inverted} {b_inverted}'

        else:
            msg = f'Invalid operator in { bracket_content }'
            raise InvalidOperatorError(msg)

        result = result[:start] + ' ' + expanded + ' ' + result[end + 1:]

    return result.strip()



from cubing_algs.move import Move


def sign_moves(old_moves: list[Move]) -> list[Move]:
    moves = []
    for move in old_moves:
        moves.append(move.to_sign)

    return moves


def unsign_moves(old_moves: list[Move]) -> list[Move]:
    moves = []
    for move in old_moves:
        moves.append(move.to_standard)

    return moves



from cubing_algs.constants import MAX_ITERATIONS
from cubing_algs.move import Move


def optimize_repeat_three_moves(
        old_moves: list[Move],
        max_depth: int = MAX_ITERATIONS,
) -> list[Move]:
    """
    R, R, R --> R'
    """
    if max_depth <= 0:
        return old_moves

    i = 0
    changed = False
    moves = old_moves.copy()

    while i < len(moves) - 2:
        if (
            not moves[i].is_pause
            and moves[i].untimed == moves[i + 1].untimed == moves[i + 2].untimed
        ):
            moves[i:i + 3] = [moves[i + 2].inverted]
            changed = True
        else:
            i += 1

    if changed:
        return optimize_repeat_three_moves(moves, max_depth - 1)

    return moves


def optimize_do_undo_moves(
        old_moves: list[Move],
        max_depth: int = MAX_ITERATIONS,
) -> list[Move]:
    """
    R R' --> <nothing>
    R2 R2 --> <nothing>
    R R R' R' --> <nothing>
    """
    if max_depth <= 0:
        return old_moves

    i = 0
    changed = False
    moves = old_moves.copy()

    while i < len(moves) - 1:
        if (
            (not moves[i].is_pause
             and moves[i].inverted.untimed == moves[i + 1].untimed) or (
                 moves[i].untimed == moves[i + 1].untimed
                 and moves[i].is_double
                 and not moves[i].is_pause
             )
        ):
            moves[i:i + 2] = []
            changed = True
        else:
            i += 1

    if changed:
        return optimize_do_undo_moves(moves, max_depth - 1)

    return moves


def optimize_double_moves(
        old_moves: list[Move],
        max_depth: int = MAX_ITERATIONS,
) -> list[Move]:
    """
    R, R --> R2
    """
    if max_depth <= 0:
        return old_moves

    i = 0
    changed = False
    moves = old_moves.copy()

    while i < len(moves) - 1:
        if (
            not moves[i].is_pause
            and not moves[i].is_double
            and moves[i].untimed == moves[i + 1].untimed
        ):
            moves[i:i + 2] = [moves[i + 1].doubled]
            changed = True
        else:
            i += 1

    if changed:
        return optimize_double_moves(moves, max_depth - 1)

    return moves


def optimize_triple_moves(
        old_moves: list[Move],
        max_depth: int = MAX_ITERATIONS,
) -> list[Move]:
    """
    R, R2 --> R'
    R2, R --> R'
    R', R2 --> R
    """
    if max_depth <= 0:
        return old_moves

    i = 0
    changed = False
    moves = old_moves.copy()

    while i < len(moves) - 1:
        if moves[i].base_move == moves[i + 1].base_move:
            if moves[i].is_double and not moves[i + 1].is_double:
                moves[i:i + 2] = [moves[i + 1].inverted]
                changed = True
            elif not moves[i].is_double and moves[i + 1].is_double:
                moves[i:i + 2] = [moves[i].inverted]
                changed = True
            else:
                i += 1
        else:
            i += 1

    if changed:
        return optimize_triple_moves(moves, max_depth - 1)

    return moves



from collections import UserList
from collections.abc import Callable
from collections.abc import Iterable
from functools import cached_property

from cubing_algs.constants import MAX_ITERATIONS
from cubing_algs.exceptions import InvalidMoveError
from cubing_algs.metrics import compute_metrics
from cubing_algs.move import Move


class Algorithm(UserList[Move]):
    """
    Represents a sequence of Rubik's cube moves.

    This class encapsulates a series of moves to be applied to a Rubik's cube,
    providing methods to manipulate and analyze the algorithm.
    """

    def __init__(self, initlist: Iterable[Move] | None = None):
        super().__init__()

        if initlist is not None:
            self.data.extend(initlist)

    @staticmethod
    def parse_moves(items: str | list[str]) -> 'Algorithm':
        from cubing_algs.parsing import parse_moves  # noqa: PLC0415

        return parse_moves(items, secure=False)

    @staticmethod
    def parse_move(item: str) -> Move:
        move = Move(item)
        if not move.is_valid:
            msg = f'{ item } is an invalid move'
            raise InvalidMoveError(msg)

        return move

    def append(self, item) -> None:
        self.data.append(self.parse_move(item))

    def insert(self, i, item) -> None:
        self.data.insert(i, self.parse_move(item))

    def extend(self, other) -> None:
        if isinstance(other, Algorithm):
            self.data.extend(other)
        else:
            self.data.extend(self.parse_moves(other))

    def __iadd__(self, other) -> 'Algorithm':
        self.extend(other)
        return self

    def __radd__(self, other) -> 'Algorithm':
        result = self.parse_moves(other)
        result += self
        return result

    def __add__(self, other) -> 'Algorithm':
        if isinstance(other, Algorithm):
            result = self.copy()
            result.extend(other)
            return result

        result = self.copy()
        result.extend(self.parse_moves(other))
        return result

    def __setitem__(self, i, item) -> None:
        if isinstance(item, Move):
            self.data[i] = item
        else:
            self.data[i] = self.parse_moves(item)

    def __str__(self) -> str:
        """
        Convert the algorithm to a human-readable string.
        """
        return ' '.join([str(m) for m in self])

    def __repr__(self) -> str:
        """
        Return a string representation that can be used
        to recreate the algorithm.
        """
        return f'Algorithm("{ "".join([str(m) for m in self]) }")'

    def __hash__(self) -> int:
        """
        Calculate a hash value for this algorithm.

        This allows Algorithm objects to be used as dictionary keys.
        """
        return hash(self.__str__())

    @property
    def metrics(self) -> dict[str, int | list[str]]:
        """
        Calculate various metrics for this algorithm.

        Uses the compute_metrics function to analyze the algorithm's efficiency,
        move types, and other characteristics.
        """
        return compute_metrics(self)

    def transform(
            self,
            *processes: Callable[[list[Move]], list[Move]],
            to_fixpoint: bool = False,
    ) -> 'Algorithm':
        """
        Apply a series of transformation functions to the algorithm's moves.

        This method enables chaining multiple transformations together, such as
        simplification, optimization, or conversion between notations.
        """
        new_moves = self.copy()
        mod_moves = self.copy()

        max_iterations = 1
        if to_fixpoint:
            max_iterations = MAX_ITERATIONS

        for _ in range(max_iterations):
            for process in processes:
                mod_moves = process(mod_moves)

            if new_moves == mod_moves:
                break
            new_moves = mod_moves

        return Algorithm(mod_moves)

    @cached_property
    def min_cube_size(self) -> int:
        """
        Compute the minimum cube size required to apply the algorithm.

        This allows validation of the algorithm against a cube.
        """
        min_cube = 2

        for m in self:
            if m.is_layered or m.is_inner_move:
                cube = 3

                max_layers = max(m.layers)
                if max_layers > 1:
                    cube = (max_layers + 1) * 2

                min_cube = max(cube, min_cube)

        return min_cube

    @cached_property
    def is_standard(self) -> bool:
        """
        Check if algorithm is in standard notations
        """
        return not self.is_sign

    @cached_property
    def is_sign(self) -> bool:
        """
        Check if algorithm contains SiGN notations
        """
        return any(m.is_sign_move for m in self)

    @cached_property
    def has_rotations(self) -> bool:
        """
        Check if algorithm contains rotations
        """
        return any(m.is_wide_move or m.is_rotation_move for m in self)

    @cached_property
    def has_internal_rotations(self) -> bool:
        """
        Check if algorithm contains internal rotations
        inducted by the moves.
        """
        return any(m.is_wide_move for m in self)



from collections.abc import Callable
from typing import Any

from cubing_algs.move import Move
from cubing_algs.transform.offset import offset_x2_moves
from cubing_algs.transform.offset import offset_x_moves
from cubing_algs.transform.offset import offset_xprime_moves
from cubing_algs.transform.offset import offset_y2_moves
from cubing_algs.transform.offset import offset_y_moves
from cubing_algs.transform.offset import offset_yprime_moves
from cubing_algs.transform.offset import offset_z2_moves
from cubing_algs.transform.offset import offset_z_moves
from cubing_algs.transform.offset import offset_zprime_moves

DEGRIP_X: dict[str, Callable[[list[Move]], list[Move]]] = {
    'x': offset_xprime_moves,
    'x2': offset_x2_moves,
    "x'": offset_x_moves,
}

DEGRIP_Y: dict[str, Callable[[list[Move]], list[Move]]] = {
    'y': offset_yprime_moves,
    'y2': offset_y2_moves,
    "y'": offset_y_moves,
}

DEGRIP_Z: dict[str, Callable[[list[Move]], list[Move]]] = {
    'z': offset_zprime_moves,
    'z2': offset_z2_moves,
    "z'": offset_z_moves,
}


DEGRIP_FULL = {}
DEGRIP_FULL.update(DEGRIP_X)
DEGRIP_FULL.update(DEGRIP_Y)
DEGRIP_FULL.update(DEGRIP_Z)


def has_grip(
        old_moves: list[Move],
        config: dict[str, Callable[[list[Move]], list[Move]]],
) -> tuple[bool, Any, Any, Any]:
    i = 0
    prefix: list[Move] = []
    suffix: list[Move] = []

    while i < len(old_moves) - 1:
        move = old_moves[i].untimed

        if move in config:
            suffix = old_moves[i + 1:]
            prefix = old_moves[:i]
            break

        i += 1

    config_keys = set(config.keys())
    if suffix and any(move not in config_keys for move in suffix):
        return True, prefix, suffix, move

    return False, False, False, False


def degrip(
        old_moves: list[Move],
        config: dict[str, Callable[[list[Move]], list[Move]]],
) -> list[Move]:
    _gripped, prefix, suffix, gripper = has_grip(old_moves, config)

    if suffix:
        degripped = [*config[gripper](suffix), gripper]

        if has_grip(degripped, config)[0]:
            return degrip(prefix + degripped, config)

        return prefix + degripped

    return old_moves


def degrip_x_moves(old_moves: list[Move]) -> list[Move]:
    return degrip(
        old_moves, DEGRIP_X,
    )


def degrip_y_moves(old_moves: list[Move]) -> list[Move]:
    return degrip(
        old_moves, DEGRIP_Y,
    )


def degrip_z_moves(old_moves: list[Move]) -> list[Move]:
    return degrip(
        old_moves, DEGRIP_Z,
    )


def degrip_full_moves(old_moves: list[Move]) -> list[Move]:
    return degrip(
        old_moves, DEGRIP_FULL,
    )



import unittest

from cubing_algs.move import Move
from cubing_algs.parsing import parse_moves
from cubing_algs.transform.symmetry import symmetry_c_moves
from cubing_algs.transform.symmetry import symmetry_e_moves
from cubing_algs.transform.symmetry import symmetry_m_moves
from cubing_algs.transform.symmetry import symmetry_s_moves


class TransformSymmetryTestCase(unittest.TestCase):

    def test_symmetry_c_moves(self):
        provide = parse_moves("U R U' R'")
        expect = parse_moves("U L U' L'")

        result = symmetry_c_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

        provide = parse_moves("F R' U2")
        expect = parse_moves("B L' U2")

        self.assertEqual(
            symmetry_c_moves(provide),
            expect,
        )

    def test_symmetry_e_moves(self):
        provide = parse_moves("U R U' R'")
        expect = parse_moves("D' R' D R")

        self.assertEqual(
            symmetry_e_moves(provide),
            expect,
        )

        provide = parse_moves("F R' U2")
        expect = parse_moves("F' R D2")

        self.assertEqual(
            symmetry_e_moves(provide),
            expect,
        )

    def test_symmetry_m_moves(self):
        provide = parse_moves("U R U' R'")
        expect = parse_moves("U' L' U L")

        self.assertEqual(
            symmetry_m_moves(provide),
            expect,
        )

        provide = parse_moves("F R' U2")
        expect = parse_moves("F' L U2")

        self.assertEqual(
            symmetry_m_moves(provide),
            expect,
        )

        provide = parse_moves("F R' U2 M'")
        expect = parse_moves("F' L U2 M'")

        self.assertEqual(
            symmetry_m_moves(provide),
            expect,
        )

    def test_symmetry_m_moves_sign(self):
        provide = parse_moves("U R u' r'")
        expect = parse_moves("U' L' u l")

        self.assertEqual(
            symmetry_m_moves(provide),
            expect,
        )

    def test_symmetry_m_moves_wide_standard(self):
        provide = parse_moves("U R Uw' Rw'")
        expect = parse_moves("U' L' Uw Lw")

        self.assertEqual(
            symmetry_m_moves(provide),
            expect,
        )

    def test_symmetry_s_moves(self):
        provide = parse_moves("U R U' R'")
        expect = parse_moves("U' R' U R")

        self.assertEqual(
            symmetry_s_moves(provide),
            expect,
        )

        provide = parse_moves("F R' U2")
        expect = parse_moves("B' R U2")

        self.assertEqual(
            symmetry_s_moves(provide),
            expect,
        )

    def test_symmetry_m_moves_big_moves(self):
        provide = parse_moves("U R 2Uw' 3Rw'")
        expect = parse_moves("U' L' 2Uw 3Lw")

        self.assertEqual(
            symmetry_m_moves(provide),
            expect,
        )

    def test_symmetry_m_moves_big_moves_timed(self):
        provide = parse_moves("U R 2Uw'@300 3Rw'")
        expect = parse_moves("U' L' 2Uw@300 3Lw")

        self.assertEqual(
            symmetry_m_moves(provide),
            expect,
        )

    def test_symmetry_m_moves_big_moves_timed_paused(self):
        provide = parse_moves("U R .@200 2Uw'@300 3Rw'")
        expect = parse_moves("U' L' .@200 2Uw@300 3Lw")

        self.assertEqual(
            symmetry_m_moves(provide),
            expect,
        )



import unittest

from cubing_algs.move import Move
from cubing_algs.parsing import parse_moves
from cubing_algs.transform.size import compress_moves
from cubing_algs.transform.size import expand_moves


class TransformCompressTestCase(unittest.TestCase):

    def test_compress_moves(self):
        provide = parse_moves(
            "U (R U2 R' U' R U' R') "
            "(R U2 R' U' R U' R') "
            "(R U2 R' U' R U' R')",
            secure=False,
        )
        expect = parse_moves("U R U2 R' U' R U R' U' R U R' U' R U' R'")

        result = compress_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_compress_big_moves(self):
        provide = parse_moves(
            "3-4Uw (R U2 R' U' R U' 2R') "
            "(2R U2 R' U' R U' R') "
            "(R U2 R' U' R U' 4R')",
            secure=False,
        )
        expect = parse_moves("3-4Uw R U2 R' U' R U R' U' R U R' U' R U' 4R'")

        result = compress_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_compress_moves_timed(self):
        provide = parse_moves(
            "U@1 (R@2 U2@3 R'@4 U'@5 R@6 U'@7 R'@8) "
            "(R@9 U2@10 R'@11 U'@12 R@13 U'@14 R'@15) "
            "(R@16 U2@17 R'@18 U'@19 R@20 U'@21 R'@22)",
            secure=False,
        )

        expect = parse_moves(
            "U@1 R@2 U2@3 R'@4 U'@5 R@6 U@7 "
            "R'@11 U'@12 R@13 U@14 "
            "R'@18 U'@19 R@20 U'@21 R'@22",
        )

        result = compress_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_compress_moves_timed_paused(self):
        provide = parse_moves(
            "U@0 .@1 (R@2 U2@3 R'@4 U'@5 R@6 U'@7 R'@8) "
            "(R@9 U2@10 R'@11 U'@12 R@13 U'@14 R'@15) "
            "(R@16 U2@17 R'@18 U'@19 R@20 U'@21 R'@22)",
            secure=False,
        )

        expect = parse_moves(
            "U@0 .@1 R@2 U2@3 R'@4 U'@5 R@6 U@7 "
            "R'@11 U'@12 R@13 U@14 "
            "R'@18 U'@19 R@20 U'@21 R'@22",
        )

        result = compress_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_compress_moves_timed_paused_middle(self):
        provide = parse_moves(
            "U@1 (R@2 U2@3 R'@4 U'@5 R@6 U'@7 R'@8) "
            "(R@9 .@9 U2@10 R'@11 U'@12 R@13 U'@14 R'@15) "
            "(R@16 U2@17 R'@18 U'@19 R@20 U'@21 R'@22)",
            secure=False,
        )

        expect = parse_moves(
            "U@1 R@2 U2@3 R'@4 U'@5 R@6 U'@7 "
            ".@9 U2@10 R'@11 U'@12 R@13 U@14 "
            "R'@18 U'@19 R@20 U'@21 R'@22",
        )

        result = compress_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_compress_pauses(self):
        provide = parse_moves(
            'U . . . U',
        )

        expect = parse_moves(
            'U . . . U',
        )

        result = compress_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_compress_pauses_timed(self):
        provide = parse_moves(
            'U@1 .@2 .@3 .@4 U@5',
        )

        expect = parse_moves(
            'U@1 .@2 .@3 .@4 U@5',
        )

        result = compress_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_compress_moves_max(self):
        provide = parse_moves(
            "U (R U2 R' U' R U' R') "
            "(R U2 R' U' R U' R') "
            "(R U2 R' U' R U' R')",
            secure=False,
        )

        result = compress_moves(provide, 0)

        self.assertEqual(
            result,
            provide,
        )


class TransformExpandTestCase(unittest.TestCase):

    def test_expand_moves(self):
        provide = parse_moves('R2 F U')
        expect = parse_moves('R R F U')

        result = expand_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_expand_big_moves(self):
        provide = parse_moves('2R2 F U')
        expect = parse_moves('2R 2R F U')

        result = expand_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_expand_timed_moves(self):
        provide = parse_moves('2R2@1 F@2 U@3')
        expect = parse_moves('2R@1 2R@1 F@2 U@3')

        result = expand_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_expand_timed_moves_paused(self):
        provide = parse_moves('2R2@1 .@2 F@3 U@4')
        expect = parse_moves('2R@1 2R@1 .@2 F@3 U@4')

        result = expand_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))



import os
from typing import TYPE_CHECKING

from cubing_algs.constants import CROSS_MASK
from cubing_algs.constants import F2L_EDGE_ORIENTATIONS
from cubing_algs.constants import F2L_FACES
from cubing_algs.constants import F2L_MASK
from cubing_algs.constants import FACE_INDEXES
from cubing_algs.constants import FACE_ORDER
from cubing_algs.constants import FULL_MASK
from cubing_algs.constants import OLL_MASK
from cubing_algs.constants import PLL_MASK

if TYPE_CHECKING:
    from cubing_algs.vcube import VCube  # pragma: no cover

DEFAULT_COLORS = [
    'white', 'red', 'green',
    'yellow', 'orange', 'blue',
]

TERM_COLORS = {
    'reset': '\x1b[0;0m',
    'hide': '\x1b[48;5;238m\x1b[38;5;252m',
    'green': '\x1b[48;5;40m\x1b[38;5;232m',
    'blue': '\x1b[48;5;21m\x1b[38;5;230m',
    'red': '\x1b[48;5;196m\x1b[38;5;232m',
    'orange': '\x1b[48;5;208m\x1b[38;5;232m',
    'yellow': '\x1b[48;5;226m\x1b[38;5;232m',
    'white': '\x1b[48;5;254m\x1b[38;5;232m',
}

FACE_COLORS = dict(
    zip(FACE_ORDER, DEFAULT_COLORS, strict=True),
)

USE_COLORS = os.environ.get('TERM') == 'xterm-256color'


class VCubeDisplay:
    facelet_size = 3

    def __init__(self, cube: 'VCube'):
        self.cube = cube
        self.cube_size = cube.size
        self.face_size = self.cube_size * self.cube_size

    @staticmethod
    def compute_mask(cube: 'VCube', mask: str) -> str:
        if not mask:
            return FULL_MASK

        new_cube = cube.__class__(mask, check=False)

        moves = ' '.join(cube.history)

        if moves:
            new_cube.rotate(moves)

        return new_cube.state

    def split_faces(self, state: str) -> list[str]:
        return [
            state[i * self.face_size: (i + 1) * self.face_size]
            for i in range(6)
        ]

    def display(self, mode: str = '', orientation: str = '') -> str:
        mask = ''
        display_method = self.display_cube
        default_orientation = ''

        if mode == 'oll':
            mask = OLL_MASK
            display_method = self.display_top_face
            default_orientation = 'D'
        elif mode == 'pll':
            mask = PLL_MASK
            display_method = self.display_top_face
            default_orientation = 'D'
        elif mode == 'f2l':
            mask = F2L_MASK

            impacted_faces = ''
            for face in F2L_FACES:
                facelets = self.cube.get_face_by_center(face)
                if face * 6 not in facelets:
                    impacted_faces += face

            selected_front_face = F2L_EDGE_ORIENTATIONS.get(impacted_faces, '')

            default_orientation = f'D{ selected_front_face }'

        elif mode == 'cross':
            mask = CROSS_MASK
            default_orientation = 'BU'

        final_orientation = orientation or default_orientation
        if final_orientation:
            cube = self.cube.oriented_copy(final_orientation, full=True)
        else:
            cube = self.cube

        faces = self.split_faces(cube.state)
        masked_faces = self.split_faces(
            self.compute_mask(cube, mask),
        )

        return display_method(faces, masked_faces)

    @staticmethod
    def display_facelet(facelet: str, mask: str = '') -> str:
        face_color = 'hide' if mask == '0' else FACE_COLORS[facelet]

        if USE_COLORS:
            return (
                f'{ TERM_COLORS[face_color]}'
                f' { facelet } '
                f'{ TERM_COLORS["reset"] }'
            )
        return f' { facelet } '

    def display_top_down_face(self, face: str, face_mask: str) -> str:
        result = ''

        for index, facelet in enumerate(face):
            if index % self.cube_size == 0:
                result += (' ' * (self.facelet_size * self.cube_size))

            result += self.display_facelet(
                facelet,
                face_mask[index],
            )

            if index % self.cube_size == self.cube_size - 1:
                result += '\n'

        return result

    def display_top_down_adjacent_facelets(self, face: str, face_mask: str,
                                           *, top: bool = False) -> str:
        result = '   '
        facelets = face[:3]
        facelets_mask = face_mask[:3]

        if top:
            facelets = facelets[::-1]
            facelets_mask = facelets_mask[::-1]

        for index, facelet in enumerate(facelets):
            result += self.display_facelet(
                facelet,
                facelets_mask[index],
            )

        result += '\n'

        return result

    def display_cube(self, faces: list[str], faces_mask: list[str]) -> str:
        middle = [
            faces[FACE_INDEXES['L']],
            faces[FACE_INDEXES['F']],
            faces[FACE_INDEXES['R']],
            faces[FACE_INDEXES['B']],
        ]
        middle_mask = [
            faces_mask[FACE_INDEXES['L']],
            faces_mask[FACE_INDEXES['F']],
            faces_mask[FACE_INDEXES['R']],
            faces_mask[FACE_INDEXES['B']],
        ]

        # Top
        result = self.display_top_down_face(
            faces[FACE_INDEXES['U']],
            faces_mask[FACE_INDEXES['U']],
        )

        # Middle
        for i in range(self.cube_size):
            for face, face_masked in zip(middle, middle_mask, strict=True):
                for j in range(self.cube_size):
                    result += self.display_facelet(
                        face[i * self.cube_size + j],
                        face_masked[i * self.cube_size + j],
                    )
            result += '\n'

        # Bottom
        result += self.display_top_down_face(
            faces[FACE_INDEXES['D']],
            faces_mask[FACE_INDEXES['D']],
        )

        return result

    def display_top_face(self, faces: list[str],
                            faces_mask: list[str]) -> str:
        result = ''

        # Top
        result = self.display_top_down_adjacent_facelets(
            faces[FACE_INDEXES['B']],
            faces_mask[FACE_INDEXES['B']],
            top=True,
        )

        # Middle
        for line in range(3):
            result += self.display_facelet(
                faces[FACE_INDEXES['L']][line],
                faces_mask[FACE_INDEXES['L']][line],
            )

            for i in range(3):
                result += self.display_facelet(
                    faces[FACE_INDEXES['U']][line * 3 + i],
                    faces_mask[FACE_INDEXES['U']][line * 3 + i],
                )

            result += self.display_facelet(
                faces[FACE_INDEXES['R']][2 - line],
                faces_mask[FACE_INDEXES['R']][2 - line],
            )

            result += '\n'

        # Bottom
        result += self.display_top_down_adjacent_facelets(
            faces[FACE_INDEXES['F']],
            faces_mask[FACE_INDEXES['F']],
            top=False,
        )

        return result



import unittest

from cubing_algs.move import Move
from cubing_algs.parsing import parse_moves
from cubing_algs.transform.pause import pause_moves
from cubing_algs.transform.pause import unpause_moves


class TransformUnpauseTestCase(unittest.TestCase):

    def test_unpause_moves(self):
        provide = parse_moves(
            "F@1 R@2 .@3 U2@4 F'@5",
        )
        expect = parse_moves("F@1 R@2 U2@4 F'@5")

        result = unpause_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_unpause_moves_untimed(self):
        provide = parse_moves(
            "F@1 R . U2 F'@4",
        )
        expect = parse_moves("F@1 R U2 F'@4")

        result = unpause_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))


class TransformPauseTestCase(unittest.TestCase):

    def test_pause_moves(self):
        provide = parse_moves(
            "F@0 R@300 U2@1300 F'@1450",
        )
        expect = parse_moves(
            "F@0 R@300 .@800 U2@1300 F'@1450",
        )

        result = pause_moves()(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_pause_moves_empty(self):
        provide = parse_moves('')

        result = pause_moves()(provide)

        self.assertEqual(
            result,
            provide,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_pause_moves_untimed(self):
        provide = parse_moves(
            "F R U2 F'",
        )
        expect = parse_moves(
            "F R U2 F'",
        )

        result = pause_moves()(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_pause_moves_untimed_partial(self):
        provide = parse_moves(
            "F@0 R@300 U2@1300 F'",
        )
        expect = parse_moves(
            "F@0 R@300 U2@1300 F'",
        )

        result = pause_moves()(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))


class TransformPauseConfigTestCase(unittest.TestCase):

    def test_pause_moves(self):
        provide = parse_moves(
            "F@0 R@300 U2@1300 F'@1450",
        )
        expect = parse_moves(
            "F@0 R@300 .@800 U2@1300 F'@1450",
        )

        result = provide.transform(pause_moves())

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_pause_moves_configured(self):
        provide = parse_moves(
            "F@0 R@300 U2@1300 F'@1450",
        )
        expect = parse_moves(
            "F@0 R@300 .@800 U2@1300 F'@1450",
        )

        result = provide.transform(pause_moves(300, 2))

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_pause_moves_configured_multiple(self):
        provide = parse_moves(
            "F@0 R@300 U2@1300 F'@1450",
        )
        expect = parse_moves(
            "F@0 R@300 .@800 .@1200 U2@1300 F'@1450",
        )

        result = provide.transform(pause_moves(200, 2, multiple=True))

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))



"""
Collection of cool patterns on 3x3x3

https://ruwix.com/the-rubiks-cube/rubiks-cube-patterns-algorithms/more-rubiks-patterns/
"""
from cubing_algs.algorithm import Algorithm
from cubing_algs.parsing import parse_moves

PATTERNS = {
    '3T': "B U2 L2 F2 R2 F D2 F2 R2 F' R2 U2",
    '4Crosses': 'F2 B2 R F2 B2 R F2 B2 R F2 B2 R F2 B2 R F2 B2 R',
    '4Plus2Dots': "F U2 D2 R L U' D F B R U2 R2 U2 F2 L2 U2 F2 L2 B2",
    'Anaconda': "L U B' U' R L' B R' F B' D R D' F'",
    'AreYouHigh': "L R' U2 D2 L' R U2 D2 R2 L2",
    'BlackMamba': "R D L F' R L' D R' U D' B U' R' D'",
    'CUAround': "U' B2 U L2 D L2 R2 D' B' R D' L R' B2 U2 F' L' U'",
    'Cage': "L U F2 R L' U2 B' U D B2 L F B' R' L F' R",
    'CheckerZigzag': 'R2 L2 F2 B2 U F2 B2 U2 F2 B2 U',
    'CornerPyramid': "U' D B R' F R B' L' F' B L F R' B' R F' U' D",
    'Cross': "R2 L' D F2 R' D' R' L U' D R D B2 R' U D2",
    'CrossingSnake': "R L U2 R L' U2 F2 R2 D2 B2 D2 B2 L2 D2 R2",
    'CubeInACubeInACube': "U' L' U' F' R2 B' R F U B2 U B' L U' F U R F'",
    'CubeInTheCube': "F L F U' R U F2 L2 U' L' B D' B' L2 U",
    'DontCrossLine': 'F2 L2 R2 B2 U2 D2',
    'Doubler': "R L U2 R L' D2 R2 F2 U2 F2 U2 L2 F2 U2 L2",
    'EasyCheckerboard': 'U2 D2 R2 L2 F2 B2',
    'EdgeTriangle': "U B2 U' F' U' D L' D2 L U D' F D' L2 B2 D'",
    'ExchangedChickenFeet': "F L' D' B' L F U F' D' F L2 B' R' U L2 D' F",
    'ExchangedDuckFeet': "U F R2 F' D' R U B2 U2 F' R2 F D B2 R B'",
    'ExchangedPeaks': "F U2 L F L' B L U B' R' L' U R' D' F' B R2",
    'ExchangedRings': "B' U' B' L' D B U D2 B U L D' L' U' L2 D",
    'FacingCheckerboards': 'U2 F2 U2 F2 B2 U2 F2 D2',
    'FourSpots': "F2 B2 U D' R2 L2 U D'",
    'GreenMamba': "R D R F R' F' B D R' U' B' U D2",
    'Headlights': 'U2 F2 U2 D2 B2 D2',
    'HenrysSnake': "R2 F2 U' D' B2 L2 F2 L2 U D R2 F2",
    'Hi': 'R2 L2 D2 R2 L2 U2',
    'HiAllAround': 'U2 R2 F2 U2 D2 F2 L2 U2',
    'Kilt': "U' R2 L2 F2 B2 U' R L F B' U F2 D2 R2 L2 F2 U2 F2 U' F2",
    'LooseStrap': 'F2 U2 B2 U2 F2 U2',
    'MatchingPictures': "R' D2 R L D2 L'",
    'Mirror': "U D F2 B2 U' D' R2 L2 B2",
    'OppositeCheckerboards': "U D R2 L2 U D' R2 L2 D2",
    'OppositeCorners': 'R L U2 F2 D2 F2 R L F2 D2 B2 D2',
    'OrderInChaos': "B L2 B' U2 B F2 U L U B U' R U' B F U' R D R B' U'",
    'PercentSign': "R L U2 R L' U2 F2 L2 U2 F2 U2 F2 R2 U2 R2",
    'Picnic': 'D2 R2 L2 F2 B2',
    'Pillars': 'L2 R2 B2 F2',
    'PlusMinus': 'U2 R2 L2 U2 R2 L2',
    'PlusMinusCheck': 'U D R2 L2 U D R2 L2',
    'Pyraminx': "D L' U R' B' R B U2 D B D' B' L U D'",
    'Python': "F2 R' B' U R' L F' L F' B D' R B L2",
    'QuickMaths': "R L F2 U2 R' L' F2 U2 R2 F2 U R2 L2 F2 B2 D'",
    'Quote': 'U2 F2 D2 B2 R2 U2 B2 R2 U2 R2',
    'Rockets': "U R2 F2 R2 U' D F2 R2 F2 D",
    'RonsCubeInACube': "F D' F' R D F' R' D R D L' F L D R' F D'",
    'SixSpots': "U D' R L' F B' U D'",
    'SixTwoOne': "U B2 D2 L B' L' U' L' B D2 B2",
    'Slash': 'R L F B R L F B R L F B',
    'Solved': '',
    'SpeedsolvingLogo': "R' L' U2 F2 D2 F2 R L B2 U2 B2 U2",
    'Spiral': "L' B' D U R U' R' D2 R2 D L D' L' R' F U",
    'StripeDotSolved': "D U B2 F2 D' U'",
    'Superflip': "U R2 F B R B2 R U2 L B2 R U' D' R2 F R' L B2 U2 F2",
    'Tablecloth': "R L U2 F' U2 D2 R2 L2 F' D2 F2 D R2 L2 F2 B2 D B2 L2",
    'Tetris': "L R F B U' D' L' R'",
    'TwistedChickenFeet': "F L' D F' U' B U F U' F R' F2 L U' R' D2",
    'TwistedCorners': "F U D R2 F2 U' B U B U R2 B' U L2 U B2 U2 F2 L2 U'",
    'TwistedDuckFeet': "F R' B R U F' L' F' U2 L' U' D2 B D' F B' U2",
    'TwistedPeaks': "F B' U F U F U L B L2 B' U F' L U L' B",
    'TwistedRings': "F D F' D2 L' B' U L D R U L' F' U L U2",
    'Twister': "F R' U L F' L' F U' R U L' U' L F'",
    'UnionJack': "U F B' L2 U2 L2 F' B U2 L2 U",
    'VerticalStripes': "F U F R L2 B D' R D2 L D' B R2 L F U F",
    'Wire': 'R L F B R L F B R L F B R2 B2 L2 R2 B2 L2',
    'YanYing': "L R F B U' D' L' R'",
    'YinYang': "R L B F R L U' D' F' B' U D",
    'ZZLine': "R L U2 R L' U2 F2 R2 U2 F2 D2 B2 L2 U2 L2",
}


def get_pattern(pattern_name: str) -> Algorithm:
    return parse_moves(PATTERNS.get(pattern_name, ''))



from collections.abc import Callable

from cubing_algs.constants import MAX_ITERATIONS
from cubing_algs.constants import RESLICE_E_MOVES
from cubing_algs.constants import RESLICE_M_MOVES
from cubing_algs.constants import RESLICE_MOVES
from cubing_algs.constants import RESLICE_S_MOVES
from cubing_algs.constants import RESLICE_THRESHOLD
from cubing_algs.constants import UNSLICE_ROTATION_MOVES
from cubing_algs.constants import UNSLICE_WIDE_MOVES
from cubing_algs.move import Move


def unslice(old_moves: list[Move], config: dict[str, list[str]]) -> list[Move]:
    moves = []

    move_cache: dict[Move, list[Move]] = {}
    for move_str, replacements in config.items():
        move_cache[Move(move_str)] = [Move(m) for m in replacements]

    for move in old_moves:
        move_untimed = move.untimed

        if move_untimed in config:
            if move.is_timed:
                moves.extend(
                    [
                        Move(x + move.time)
                        for x in move_cache[move_untimed]
                    ],
                )
            else:
                moves.extend(move_cache[move_untimed])
        else:
            moves.append(move)

    return moves


def unslice_wide_moves(old_moves: list[Move]) -> list[Move]:
    return unslice(old_moves, UNSLICE_WIDE_MOVES)


def unslice_rotation_moves(old_moves: list[Move]) -> list[Move]:
    return unslice(old_moves, UNSLICE_ROTATION_MOVES)


def reslice(
        old_moves: list[Move],
        config: dict[str, list[str]],
        max_depth: int = MAX_ITERATIONS,
        threshold: int = 0,
) -> list[Move]:
    if max_depth <= 0:
        return old_moves

    i = 0
    moves = []
    changed = False

    while i < len(old_moves) - 1:
        current_move = old_moves[i]
        next_move = old_moves[i + 1]

        sliced = f'{ current_move.untimed } { next_move.untimed }'
        valid_threshold = True
        if (
                threshold
                and current_move.timed
                and next_move.timed
                and next_move.timed - current_move.timed > threshold
        ):
            valid_threshold = False

        if valid_threshold and sliced in config:
            for move in config[sliced]:
                moves.append(Move(move + old_moves[i].time))
            changed = True
            i += 2
        else:
            moves.append(old_moves[i])
            i += 1

    if i < len(old_moves):
        moves.append(old_moves[i])

    if changed:
        return reslice(moves, config, max_depth - 1, threshold)

    return moves


def reslice_m_moves(old_moves: list[Move]) -> list[Move]:
    return reslice(old_moves, RESLICE_M_MOVES)


def reslice_s_moves(old_moves: list[Move]) -> list[Move]:
    return reslice(old_moves, RESLICE_S_MOVES)


def reslice_e_moves(old_moves: list[Move]) -> list[Move]:
    return reslice(old_moves, RESLICE_E_MOVES)


def reslice_moves(old_moves: list[Move]) -> list[Move]:
    return reslice(old_moves, RESLICE_MOVES)


def reslice_m_timed_moves(
        threshold: int = RESLICE_THRESHOLD,
) -> Callable[[list[Move]], list[Move]]:

    def _reslice_timed_moves(old_moves: list[Move]) -> list[Move]:
        return reslice(old_moves, RESLICE_M_MOVES, threshold=threshold)

    return _reslice_timed_moves


def reslice_s_timed_moves(
        threshold: int = RESLICE_THRESHOLD,
) -> Callable[[list[Move]], list[Move]]:

    def _reslice_timed_moves(old_moves: list[Move]) -> list[Move]:
        return reslice(old_moves, RESLICE_S_MOVES, threshold=threshold)

    return _reslice_timed_moves


def reslice_e_timed_moves(
        threshold: int = RESLICE_THRESHOLD,
) -> Callable[[list[Move]], list[Move]]:
    def _reslice_timed_moves(old_moves: list[Move]) -> list[Move]:
        return reslice(old_moves, RESLICE_E_MOVES, threshold=threshold)

    return _reslice_timed_moves


def reslice_timed_moves(
        threshold: int = RESLICE_THRESHOLD,
) -> Callable[[list[Move]], list[Move]]:

    def _reslice_timed_moves(old_moves: list[Move]) -> list[Move]:
        return reslice(old_moves, RESLICE_MOVES, threshold=threshold)

    return _reslice_timed_moves



# ruff: noqa: PLR6104 RUF005
import unittest

from cubing_algs.algorithm import Algorithm
from cubing_algs.exceptions import InvalidMoveError
from cubing_algs.move import Move
from cubing_algs.parsing import parse_moves
from cubing_algs.transform.optimize import optimize_do_undo_moves
from cubing_algs.transform.optimize import optimize_double_moves


class AlgorithmTestCase(unittest.TestCase):

    def test_init_empty(self):
        algo = Algorithm()
        self.assertEqual(str(algo), '')

        algo.extend('R2 U')

        self.assertEqual(str(algo), 'R2 U')

    def test_parse_moves(self):
        algo = Algorithm.parse_moves('R2 U')
        self.assertEqual(str(algo), 'R2 U')

        algo = Algorithm.parse_moves(['R2', 'U'])
        self.assertEqual(str(algo), 'R2 U')

        algo = Algorithm.parse_moves(Algorithm.parse_moves(['R2', 'U']))
        self.assertEqual(str(algo), 'R2 U')

    def test_parse_move(self):
        move = Algorithm.parse_move('R2')
        self.assertEqual(move, 'R2')

        with self.assertRaises(InvalidMoveError):
            Algorithm.parse_move('R2 U')

    def test_append(self):
        algo = parse_moves('R2 U')
        self.assertEqual(str(algo), 'R2 U')
        algo.append('F2')
        self.assertEqual(str(algo), 'R2 U F2')

        for m in algo:
            self.assertTrue(isinstance(m, Move))

        algo.append(Move('D'))
        self.assertEqual(str(algo), 'R2 U F2 D')

        for m in algo:
            self.assertTrue(isinstance(m, Move))

        with self.assertRaises(InvalidMoveError):
            algo.append('G')

        with self.assertRaises(InvalidMoveError):
            algo.append('F R')

        with self.assertRaises(InvalidMoveError):
            algo.append(Move('G'))

        with self.assertRaises(InvalidMoveError):
            algo.append(['F', 'R'])

        with self.assertRaises(InvalidMoveError):
            algo.append([])

    def test_extend(self):
        algo = parse_moves('R2 U')
        self.assertEqual(str(algo), 'R2 U')
        algo.extend(['F2', 'B'])
        self.assertEqual(str(algo), 'R2 U F2 B')

        for m in algo:
            self.assertTrue(isinstance(m, Move))

        algo.extend([])
        self.assertEqual(str(algo), 'R2 U F2 B')

        for m in algo:
            self.assertTrue(isinstance(m, Move))

        algo.extend('F R')
        self.assertEqual(str(algo), 'R2 U F2 B F R')

        with self.assertRaises(InvalidMoveError):
            algo.extend(['F2', 'G'])

    def test_extend_with_algorithm(self):
        algo = parse_moves('R2 U')
        self.assertEqual(str(algo), 'R2 U')
        algo.extend(parse_moves('F2 B'))
        self.assertEqual(str(algo), 'R2 U F2 B')

        for m in algo:
            self.assertTrue(isinstance(m, Move))

    def test_add_operator(self):
        algo = parse_moves('R2 U')
        self.assertEqual(str(algo), 'R2 U')
        algo = algo + ['F2', 'B']
        self.assertEqual(str(algo), 'R2 U F2 B')

        for m in algo:
            self.assertTrue(isinstance(m, Move))

        algo = algo + 'F R'
        self.assertEqual(str(algo), 'R2 U F2 B F R')

        for m in algo:
            self.assertTrue(isinstance(m, Move))

        with self.assertRaises(InvalidMoveError):
            algo += 'F2 G'

    def test_iadd_operator(self):
        algo = parse_moves('R2 U')
        self.assertEqual(str(algo), 'R2 U')
        algo += ['F2', 'B']
        self.assertEqual(str(algo), 'R2 U F2 B')

        for m in algo:
            self.assertTrue(isinstance(m, Move))

        algo += 'F R'
        self.assertEqual(str(algo), 'R2 U F2 B F R')

        for m in algo:
            self.assertTrue(isinstance(m, Move))

        with self.assertRaises(InvalidMoveError):
            algo + 'F2 G'

    def test_radd_operator(self):
        algo = 'F2R2' + parse_moves('D2 U')
        self.assertEqual(str(algo), 'F2 R2 D2 U')

        for m in algo:
            self.assertTrue(isinstance(m, Move))

        with self.assertRaises(InvalidMoveError):
            'F2 G' + algo

    def test_radd_operator_z(self):
        algo = 'z' + parse_moves('R2 U')
        self.assertEqual(str(algo), 'z R2 U')

    def test_radd_operator_zprime(self):
        algo = "z'" + parse_moves('R2 U')
        self.assertEqual(str(algo), "z' R2 U")

    def test_radd_operator_z2(self):
        algo = 'z2' + parse_moves('R2 U')
        self.assertEqual(str(algo), 'z2 R2 U')

    def test_add_exploded(self):
        algo = parse_moves('R2 U')
        self.assertEqual(str(algo), 'R2 U')
        algo += [*algo, 'F2', 'B']
        self.assertEqual(str(algo), 'R2 U R2 U F2 B')

        for m in algo:
            self.assertTrue(isinstance(m, Move))

    def test_add_operator_with_algorithm(self):
        algo = parse_moves('R2 U')
        self.assertEqual(str(algo), 'R2 U')
        algo = algo + parse_moves('F2 B')
        self.assertEqual(str(algo), 'R2 U F2 B')

        for m in algo:
            self.assertTrue(isinstance(m, Move))

    def test_insert(self):
        algo = parse_moves('R2 U')
        self.assertEqual(str(algo), 'R2 U')
        algo.insert(0, 'F2')
        self.assertEqual(str(algo), 'F2 R2 U')

        for m in algo:
            self.assertTrue(isinstance(m, Move))

    def test_remove(self):
        algo = parse_moves('R2 U R2')
        self.assertEqual(str(algo), 'R2 U R2')
        algo.remove('R2')
        self.assertEqual(str(algo), 'U R2')

        for m in algo:
            self.assertTrue(isinstance(m, Move))

    def test_pop(self):
        algo = parse_moves('R2 U')
        self.assertEqual(str(algo), 'R2 U')
        popped = algo.pop()
        self.assertEqual(str(algo), 'R2')
        self.assertEqual(popped, 'U')

        for m in algo:
            self.assertTrue(isinstance(m, Move))

    def test_copy(self):
        algo = parse_moves('R2 U')
        self.assertEqual(str(algo), 'R2 U')

        copy = algo.copy()
        self.assertTrue(isinstance(copy, Algorithm))
        self.assertEqual(str(copy), 'R2 U')

        for m in algo:
            self.assertTrue(isinstance(m, Move))
        for m in copy:
            self.assertTrue(isinstance(m, Move))

        algo.pop()
        self.assertEqual(str(algo), 'R2')
        self.assertEqual(str(copy), 'R2 U')

        for m in algo:
            self.assertTrue(isinstance(m, Move))
        for m in copy:
            self.assertTrue(isinstance(m, Move))

    def test_iter(self):
        algo = parse_moves('R2 U')
        for m, n in zip(algo, ['R2', 'U'], strict=True):
            self.assertEqual(m, n)

    def test_getitem(self):
        algo = parse_moves('R2 U')
        self.assertEqual(algo[1], 'U')
        self.assertEqual(type(algo[1]), Move)

    def test_setitem_slice(self):
        algo = parse_moves('R2 U F')
        algo[2:] = []
        self.assertEqual(str(algo), 'R2 U')

        for m in algo:
            self.assertTrue(isinstance(m, Move))

        algo = parse_moves('R2 U F')
        algo[1:] = []
        self.assertEqual(str(algo), 'R2')

        for m in algo:
            self.assertTrue(isinstance(m, Move))

        algo = parse_moves('R2 U F')
        new_algo = parse_moves('B2 D')
        algo[2:] = new_algo
        self.assertEqual(str(algo), 'R2 U B2 D')

        for m in algo:
            self.assertTrue(isinstance(m, Move))

        algo = parse_moves('R2 U F')
        new_algo = parse_moves('B2 D')
        algo[1:] = new_algo
        self.assertEqual(str(algo), 'R2 B2 D')

        for m in algo:
            self.assertTrue(isinstance(m, Move))

        algo = parse_moves('R2 U F L D')
        new_algo = parse_moves('B2 D')
        algo[1:3] = new_algo
        self.assertEqual(str(algo), 'R2 B2 D L D')

        for m in algo:
            self.assertTrue(isinstance(m, Move))

    def test_setitem(self):
        algo = parse_moves('R2 U')
        algo[1] = Move('B')
        self.assertEqual(str(algo), 'R2 B')

        for m in algo:
            self.assertTrue(isinstance(m, Move))

    def test_delitem(self):
        algo = parse_moves('R2 U')
        del algo[1]
        self.assertEqual(str(algo), 'R2')

        for m in algo:
            self.assertTrue(isinstance(m, Move))

    def test_length(self):
        algo = parse_moves('R2 U')

        self.assertEqual(len(algo), 2)

    def test_str(self):
        algo = parse_moves('R2 U')

        self.assertEqual(str(algo), 'R2 U')

    def test_repr(self):
        algo = parse_moves('R2 U')

        self.assertEqual(repr(algo), 'Algorithm("R2U")')

    def test_eq(self):
        algo = parse_moves('R2 U')
        algo_bis = parse_moves('R2 U')

        self.assertEqual(algo, algo_bis)

    def test_eq_copy(self):
        algo = parse_moves('R2 U')
        algo_copy = algo.copy()

        self.assertEqual(algo, algo_copy)

    def test_eq_list(self):
        algo = parse_moves('R2 U')
        algo_list = [Move('R2'), Move('U')]

        self.assertEqual(algo, algo_list)

    def test_hash(self):
        algo = parse_moves('R2 U')
        self.assertTrue(hash(algo))

    def test_transform(self):
        algo = parse_moves('R R U F2 F2')
        expected = parse_moves('R2 U')

        self.assertEqual(
            algo.transform(
                optimize_do_undo_moves,
                optimize_double_moves,
            ),
            expected,
        )

        algo = parse_moves('R U F2')
        expected = parse_moves('R U F2')

        self.assertEqual(
            algo.transform(
                optimize_do_undo_moves,
                optimize_double_moves,
            ),
            expected,
        )

    def test_transform_to_fixpoint(self):
        algo = parse_moves("R R F F' R2 U F2")
        expected = parse_moves('R2 R2 U F2')

        self.assertEqual(
            algo.transform(
                optimize_do_undo_moves,
                optimize_double_moves,
            ),
            expected,
        )

        algo = parse_moves("R R F F' R2 U F2")
        expected = parse_moves('U F2')

        self.assertEqual(
            algo.transform(
                optimize_do_undo_moves,
                optimize_double_moves,
                to_fixpoint=True,
            ),
            expected,
        )

    def test_min_cube_size(self):
        algo = parse_moves("B' R2 U F2")

        self.assertEqual(
            algo.min_cube_size,
            2,
        )

        algo = parse_moves("B' R2 M U F2")

        self.assertEqual(
            algo.min_cube_size,
            3,
        )

        algo = parse_moves("B' r2 U F2")

        self.assertEqual(
            algo.min_cube_size,
            3,
        )

        algo = parse_moves("B' R2 U Fw2")

        self.assertEqual(
            algo.min_cube_size,
            3,
        )

        algo = parse_moves("B' R2 U 2Fw2")

        self.assertEqual(
            algo.min_cube_size,
            3,
        )

        algo = parse_moves("B' R2 U 3Fw2")

        self.assertEqual(
            algo.min_cube_size,
            6,
        )

        algo = parse_moves("B' R2 U 2-3Fw2")

        self.assertEqual(
            algo.min_cube_size,
            6,
        )

        algo = parse_moves("B' R2 U 4Fw2")

        self.assertEqual(
            algo.min_cube_size,
            8,
        )

        algo = parse_moves("B' R2 U 2-4Fw2")

        self.assertEqual(
            algo.min_cube_size,
            8,
        )

    def test_is_standard(self):
        algo = parse_moves("B' R2 U 2-4Fw2")

        self.assertTrue(algo.is_standard)
        self.assertFalse(algo.is_sign)

    def test_is_sign(self):
        algo = parse_moves("B' R2 U 2-4f2")

        self.assertTrue(algo.is_sign)
        self.assertFalse(algo.is_standard)

    def test_has_rotations(self):
        algo = parse_moves("B' R2 U 2-4Fw2")

        self.assertTrue(algo.has_rotations)

        algo = parse_moves("B' R2 U 2-4f2")

        self.assertTrue(algo.has_rotations)

        algo = parse_moves("B' R2 U x")

        self.assertTrue(algo.has_rotations)

    def test_has_internal_rotations(self):
        algo = parse_moves("B' R2 U 2-4Fw2")

        self.assertTrue(algo.has_internal_rotations)

        algo = parse_moves("B' R2 U 2-4f2")

        self.assertTrue(algo.has_internal_rotations)

        algo = parse_moves("B' R2 U x")

        self.assertFalse(algo.has_internal_rotations)
